from datetime import datetime
import os
from threading import Lock
from loguru import logger
import numpy as np

from qlsdk.core.entity import RscPacket
from qlsdk.persist.rsc_edf import RscEDFHandler
from qlsdk.persist.stream import EDF_FILE_TYPE, EDFStreamWriter


def intersection_positions(A, B):
    setB = set(B)
    seen = set()
    return [idx for idx, elem in enumerate(A) 
            if elem in setB and elem not in seen and not seen.add(elem)]

# 用作数据结构一致化处理，通过调用公共类写入edf文件
# 入参包含写入edf的全部前置参数
# 实时数据包为个性化数据包，含有eeg数据部分
class ARSKindlingEDFHandler(object):
    '''
        Rsc EDF Handler
        处理EDF文件的读写
        RSC设备通道数根据选择变化，不同通道采样频率相同
        eeg_sample_rate: 采样频率
        physical_max: 物理最大值 (uV)   
        physical_min: 物理最小值 (uV)
        resolution: 分辨率
        storage_path: 存储路径        
        
        @author: qlsdk
        @since: 0.4.0
    '''
    def __init__(self, eeg_sample_rate, physical_max, physical_min, resolution=24, storage_path = None):
        # edf文件参数
        self.physical_max = physical_max
        self.physical_min = physical_min
        self.digital_max = 8388607 if resolution == 24 else 32767
        self.digital_min = -8388607 if resolution == 24 else - 32768
        self.file_type = EDF_FILE_TYPE["bdf"] if resolution == 24 else EDF_FILE_TYPE["edf"]
        # 点分辨率
        self.resolution = resolution
        # eeg通道数
        self.channels = None
        # eeg采样率
        self.sample_rate = eeg_sample_rate
        # bytes per second
        self.bytes_per_second = 0
        self._edf_writer = None
        self._cache2 = tuple()
        self._recording = False
        self._edf_writer = None
        self.annotations = None
        # 每个数据块大小
        self._chunk = np.array([])
        self._duration = 0
        self._points = 0
        self._first_pkg_id = None
        self._last_pkg_id = None
        self._first_timestamp = None
        self._start_time = None
        self._end_time = None
        self._patient_code = "patient_code"
        self._patient_name = "patient_name"
        self._device_type = "24130032"
        self._device_no = "24130032"
        self._total_packets = 0
        self._lost_packets = 0
        self._storage_path = storage_path
        self._edf_writer_thread = None
        self._file_prefix = None
        
        # 癫痫造模仪1对4， 这个是固定模式，写死，规则随设备
        self._edf_handler = {}
        self._edf_Handler_A = None
        self._edf_Handler_B = None  
        self._edf_Handler_C = None
        self._edf_Handler_D = None
        
        self._channel_spilt = {
            "A" : [49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62],
            "B" : [33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46],
            "C" : [17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30],
            "D" : [7, 8, 5, 6, 3, 4, 1, 2, 9, 10, 11, 12, 13, 14],
        }
        
        # self._channel_mapping = {
        #     1: "D7", 
        #     2: "D8",
        #     3: "D5", 
        #     4: "D6",
        #     5: "D3", 
        #     6: "D4",
        #     7: "D1", 
        #     8: "D2",
        #     9: "D9", 
        #     10: "D10",
        #     11: "D11", 
        #     12: "D12",
        #     13: "D13", 
        #     14: "D14",
            
        #     17: "C7", 
        #     18: "C8",
        #     19: "C5", 
        #     20: "C6",
        #     21: "C3", 
        #     22: "C4",
        #     23: "C1", 
        #     24: "C2",
        #     25: "C9", 
        #     26: "C10",
        #     27: "C11", 
        #     28: "C12",
        #     29: "C13", 
        #     30: "C14",
            
        #     33: "B7", 
        #     34: "B8",
        #     35: "B5", 
        #     36: "B6",
        #     37: "B3", 
        #     38: "B4",
        #     39: "B1", 
        #     40: "B2",
        #     41: "B9", 
        #     42: "B10",
        #     43: "B11", 
        #     44: "B12",
        #     45: "B13", 
        #     46: "B14",
            
        #     49: "A7", 
        #     50: "A8",
        #     51: "A5", 
        #     52: "A6",
        #     53: "A3", 
        #     54: "A4",
        #     55: "A1", 
        #     56: "A2",
        #     57: "A9", 
        #     58: "A10",
        #     59: "A11", 
        #     60: "A12",
        #     61: "A13", 
        #     62: "A14"
        # }
        
        self._lock = Lock()
        
    @property
    def file_name(self): 
        suffix = "bdf" if self.resolution == 24 else "edf"
        
        # 文件名称
        file_name = f"{self._file_prefix}_{self._device_no}_{self._start_time.strftime('%y%m%d%H%I%M')}.{suffix}" if self._file_prefix else f"{self._device_no}_{self._start_time.strftime('%y%m%d%H%I%M')}.{suffix}"
        
        if self._storage_path:
            try:
                # 自动创建目录，存在则忽略
                os.makedirs(self._storage_path, exist_ok=True)  
                
                return f"{self._storage_path}/{file_name}"
            except Exception as e:
                logger.error(f"创建目录[{self._storage_path}]失败: {e}")            
        
        return file_name
    
    def set_device_type(self, device_type):
        if device_type == 0x39:
            self._device_type = "C64RS"
        elif device_type == 0x40:
            self._device_type = "LJ64S1"
        elif device_type == 0x60:
            self._device_type = "ArsKindling"
        else:
            self._device_type = hex(device_type)
        
    def set_device_no(self, device_no):
        self._device_no = device_no
        
    def set_storage_path(self, storage_path):
        self._storage_path = storage_path
        
    def set_file_prefix(self, file_prefix):
        self._file_prefix = file_prefix
        
    def set_patient_code(self, patient_code):
        self._patient_code = patient_code
        
    def set_patient_name(self, patient_name):
        self._patient_name = patient_name
    
    def write(self, packet: RscPacket):
        # logger.trace(f"packet: {packet}")    
        if packet is None:
            # self._edf_writer_thread.stop_recording()
            for k in self._edf_handler.keys():
                self._edf_handler[k].write(None)
            return
         
        #按分区写入数据        
        for k in self._channel_spilt.keys():      
            logger.trace(f'分区{k}写入数据到文件')  
            if k in packet.channels.keys():    
                p = packet
                self.writeA(p, k)
     
    def writeA(self, packet: RscPacket, name='A'):
        # 参数检查
        if packet is None:
            logger.warning("空数据，忽略")
            return
            
        channels = packet.channels.get(name, None)
        eeg = packet.eeg.get(name, None)
        
        if eeg is None :
            logger.trace(f"分区{name}没有可用的数据，跳过文件写入")
            return
        
        # 分区数据包写入
        if name not in self._edf_handler.keys():
            edf_handler = RscEDFHandler(self.sample_rate,  self.physical_max , self.physical_min, self.resolution)  
            edf_handler.set_device_type(self._device_type)
            edf_handler.set_device_no(self._device_no)
            edf_handler.set_storage_path(self._storage_path)
            edf_handler.set_file_prefix(f'{self._file_prefix}_{name}' if self._file_prefix else name)
            logger.info(f"开始写入分区{name}的数据到文件")
            self._edf_handler[name] = edf_handler
        
        # 新建数据包实例
        data = RscPacket()
        data.time_stamp = packet.time_stamp
        data.pkg_id = packet.pkg_id
        # data.channels = channels
        data.channels = self.channel_display(channels, name)
        data.origin_sample_rate = packet.origin_sample_rate
        data.sample_rate = packet.sample_rate
        data.sample_num = packet.sample_num
        data.resolution = packet.resolution
        data.trigger = packet.trigger
        data.eeg = eeg
        
        self._edf_handler[name].write(data)
    
    # trigger标记
    # desc: 标记内容
    # cur_time: 设备时间时间戳，非设备发出的trigger不要设置
    def trigger(self, desc, cur_time=None):
        # trigger现在（20250702）多个分区共享， 分发到所有分区文件中标记
        for k in self._edf_handler.keys():
            self._edf_handler[k].trigger(desc, cur_time) 
            
    def channel_display(self, channel, name: str):
        return [f'{name}-{channel_id}' for channel_id in channel]



from enum import Enum
from bitarray import bitarray
from loguru import logger

# 通道数组[1,2,3,...]转换为bytes
def to_bytes(channels: list[int], upper=256) -> bytes:
    byte_len = int(upper / 8)
    channel = [0] * byte_len
    result = bitarray(upper)
    result.setall(0)
    for i in range(len(channels)):
        if channels[i] > 0 and channels[i] <= upper: 
            # 每个字节从低位开始计数
            m = (channels[i] - 1) % 8
            result[channels[i] + 6 - 2 * m] = 1
    return result.tobytes()

class C64Channel(Enum):
    CH0 = 0
    CH1 = 1
    CH2 = 2
    CH3 = 3
    CH4 = 4
    CH5 = 5
    CH6 = 6
    CH7 = 7
    CH8 = 8
    CH9 = 9
    CH10 = 10
    CH11 = 11
    CH12 = 12
    CH13 = 13
    CH14 = 14
    CH15 = 15

class WaveForm(Enum):
    DC = 0
    SQUARE = 1
    AC = 2
    CUSTOM = 3
    PULSE = 4

 # 刺激通道 
class StimulationChannel(object):
    def __init__(self, channel_id: int, waveform: int, current: float, duration: float, ramp_up: float = None, ramp_down: float = None,
                 frequency: float = None, phase_position: int = None, duration_delay: float = None, pulse_width: int = None, pulse_width_rate: int = 1):
        self.channel_id = channel_id
        self.waveform = waveform
        self.current_max = current
        self.current_min = current
        self.duration = duration
        self.ramp_up = ramp_up
        self.ramp_down = ramp_down
        self.frequency = frequency
        self.phase_position = phase_position
        self.duration_delay = duration_delay
        self.pulse_width = pulse_width 
        self.delay_time = 0
        self.pulse_interval = 0
        self.with_group_repeats = 1
        self.pulse_width_rate = 1065353216
        self.pulse_time_f = 0
        self.pulse_time_out = 0
        self.pulse_time_idle = 0
    
    def to_bytes(self):
        # Convert the object to bytes for transmission
        result = self.channel_id.to_bytes(1, 'little')
        wave_form = WaveForm.SQUARE.value if self.waveform == WaveForm.PULSE.value else self.waveform
        result += wave_form.to_bytes(1, 'little')
        result += int(self.current_max * 1000 * 1000).to_bytes(4, 'little')
        # result += int(self.current_min * 1000).to_bytes(2, 'little')
        result += int(self.frequency).to_bytes(2, 'little')
        result += int(self.pulse_width).to_bytes(2, 'little')
        result += int(self.pulse_width_rate).to_bytes(4, 'little')
        
        result += int(self.pulse_interval).to_bytes(2, 'little')
        result += int(self.with_group_repeats).to_bytes(2, 'little')
        result += int(self.pulse_time_f).to_bytes(4, 'little')
        result += int(self.pulse_time_out).to_bytes(4, 'little')
        result += int(self.pulse_time_idle).to_bytes(4, 'little')
        
        result += int(self.delay_time).to_bytes(4, 'little')
        result += int(self.ramp_up * 1000).to_bytes(4, 'little')
        result += int((self.duration + self.ramp_up)  * 1000).to_bytes(4, 'little')
        result += int(self.ramp_down * 1000).to_bytes(4, 'little')
        
        return result
    
    def to_json(self):
        return {
            "channel_id": self.channel_id,
            "waveform": self.waveform,
            "current_max": self.current_max,
            "current_min": self.current_min,
            "duration": self.duration,
            "ramp_up": self.ramp_up,
            "ramp_down": self.ramp_down,
            "frequency": self.frequency,
            "phase_position": self.phase_position,
            "duration_delay": self.duration_delay,
            "pulse_width": self.pulse_width,
            "delay_time": self.delay_time,
            "pulse_interval": self.pulse_interval,
            "with_group_repeats": self.with_group_repeats
        }
        
# 刺激范式
class StimulationParadigm(object):
    def __init__(self):
        self.channels = None
        self.duration = None
        self.interval_time = 0
        self.characteristic = 0
        self.mode = 0
        self.repeats = 0
        
    def add_channel(self, channel: StimulationChannel, update=False):
        if self.channels is None:
            self.channels = {}
        channel_id = channel.channel_id + 1
        if channel_id in self.channels.keys():
            logger.warning(f"Channel {channel_id} already exists")
            if update:
                self.channels[channel_id] = channel
        else:
            self.channels[channel_id] = channel  
            
        # 计算刺激时间
        duration = channel.duration + channel.ramp_up + channel.ramp_down     
        if self.duration is None or duration > self.duration:
            self.duration = duration 
        
        
    def to_bytes(self):
        result = to_bytes(list(self.channels.keys()), 64)
        result += int(self.duration * 1000).to_bytes(4, 'little')
        result += int(self.interval_time).to_bytes(4, 'little')
        result += int(self.characteristic).to_bytes(4, 'little')
        result += int(self.mode).to_bytes(1, 'little')
        result += int(self.repeats).to_bytes(4, 'little')
        for channel in self.channels.values():
            result += channel.to_bytes()
        return result
    
    def to_json(self):
        # Convert the object to JSON for transmission
        return {
            "channels": list(self.channels.keys()),
            "duration": self.duration,
            "interval_time": self.interval_time,
            "characteristic": self.characteristic,
            "mode": self.mode,
            "repeats": self.repeats, 
            "stim": [channel.to_json() for channel in self.channels.values()]
        }
        
    # @staticmethod
    # def from_json(param: Dict[str, Any]):
    #     pass
import struct    
if __name__ == "__main__":
    src = bytes.fromhex("0000a040")
    logger.info(f"src: {src}")
    result = struct.unpack("<f", src)[0]
    logger.info(f"result: {result}")
    logger.info(f"result: {struct.pack('<f', result).hex()}")




import socket
import time
from threading import Thread, Lock
from loguru import logger

from qlsdk.core.message import UDPMessage
from qlsdk.core.local import get_ips

'''
    广播器类，用于发送和接收设备广播消息
    主要功能：发送设备搜索消息，接收设备连接消息
    注意：广播端口需要和ar4sdk做区分，使用54366时不能和x8同时使用
'''
class UdpBroadcaster:
    # 广播端口需要和ar4sdk做区分， 使用54366时不能和x8同时使用
    def __init__(self, port=54366):
        self.broadcast_port = port
        self.devices_to_broadcast = []  # 待广播的设备序列号列表
        self.connected_devices = set()  # 已连接的设备序列号集合
        self.lock = Lock()  # 用于线程安全的锁
        self.running = True

        # 创建UDP socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    # 添加设备序列号到待广播列表
    def add_device(self, device_id):
        
        with self.lock:
            if device_id not in self.devices_to_broadcast:
                self.devices_to_broadcast.append(device_id)
                logger.info(f"添加设备[{device_id}]到搜索列表。")

    # 从待广播列表移除设备序列号
    def remove_device(self, device_id):
        
        with self.lock:
            if device_id in self.devices_to_broadcast:
                self.devices_to_broadcast.remove(device_id)
                logger.info(f"把设备[{device_id}]从搜索列表中移除。")

    # 把设备标记为已连接
    def mark_device_as_connected(self, device_id):
        with self.lock:
            # 如果设备已连接，则从搜索列表中移除
            if device_id in self.devices_to_broadcast:
                self.devices_to_broadcast.remove(device_id)
                
            # 添加到已连接设备集合
            self.connected_devices.add(device_id)
            
            logger.info(f"设备[{device_id}]已连接，从搜索列表中移除。")

    # 广播设备信息，寻求配对
    def broadcast_devices(self):
        
        while self.running:
            ips = get_ips()
            with self.lock:
                for device_id in self.devices_to_broadcast:
                    message = UDPMessage.search(device_id)
                    self.sock.sendto(message, ('<broadcast>', self.broadcast_port))
                    logger.debug(f"设备[{device_id}]广播消息已发送。")
                
            # 每隔1秒发送一次广播    
            time.sleep(1)  

    def start(self):
        """启动广播线程"""
        self.broadcast_thread = Thread(target=self.broadcast_devices, daemon=True)
        self.broadcast_thread.start()

    def stop(self):
        """停止广播"""
        self.running = False
        self.broadcast_thread.join()
        self.sock.close()



from enum import Enum
from time import timezone
from qlsdk.core.crc import check_crc, crc16
from qlsdk.core.local import get_ip, get_ips

from loguru import logger


class UDPCommand(Enum):
    CONNECT = 0x10
    NOTIFY = 0x09


class UDPMessage(object):
    START = "SHQuanLan"
    MAC = 'mac'
    DEV_TYPE = "dev_type"
    VERSION_CODE = "version_code"
    VERSION_NAME = "version_name"

    def __init__(self) -> None:
        self._base = None
        self._message = None

    @staticmethod
    def parse(packet, address = None):
        plen = len(packet)
        if plen < 10:
            logger.trace("message length too short.")
            return

        start = packet[:9].decode("utf-8")
        # quanlan udp message
        if start != UDPMessage.START:
            return

        if not check_crc(packet):
            logger.warn(f"数据CRC校验失败，丢弃！")
            return
        
        # message command
        cmd = int.from_bytes(packet[10:12], 'little')

        return UDPMessage._parse(cmd, packet[12:])

    @staticmethod
    def _parse(cmd, data):
        # 只解析0x09
        if cmd == UDPCommand.NOTIFY.value:
            return UDPMessage.parseDeviceInfo(data)
        else:
            logger.trace(f'不支持的消息. cmd: {hex(cmd)} dlen: {len(data)} data: {data}')
            return None
        
    @staticmethod
    def parseDeviceInfo(data):
        device_info = {}
        try:
            device_info[UDPMessage.DEV_TYPE] = hex(int.from_bytes(data[:2], 'little'))
            device_info[UDPMessage.MAC] = hex(int.from_bytes(data[2:10], 'little'))
            device_info[UDPMessage.VERSION_CODE] = int.from_bytes(data[42:46], 'little')
            device_info[UDPMessage.VERSION_NAME] = str(data[10:42],'utf-8').split('\x00')[0]
        except Exception as e:
            logger.error(f"parseDeviceInfo异常：{e}")

        return device_info

    @staticmethod
    def search(device_id : str, server_ip : str=None, server_port : int=19216):
        # 服务端Ip
        if server_ip is None:
            server_ip = get_ip()
            
        return UDPMessage._search_(device_id, server_ip, server_port)
    
    @staticmethod
    def _search_(device_id : str, server_ip : str, server_port : int=19216):
        # 服务端Ip
        if server_ip is None:
            raise ValueError("server_ip is None")
            
        logger.debug(f"search device {device_id} on {server_ip}:{server_port}")
            
        message = bytearray(28)
        # 消息头
        message[:10] = UDPMessage.START.encode('utf-8')
        # 消息类型
        message[10:12] = UDPCommand.CONNECT.value.to_bytes(2, 'little')
        # 设备序列号格式沿用现有协议
        serial_no = f'F0{device_id[:2]}FFFFFFFF{device_id[4:]}' 
        message[12:22] = bytes.fromhex(serial_no)
        # 本机ip
        ip = server_ip.split(".")    
        message[22] = (int)(ip[0])
        message[23] = (int)(ip[1])
        message[24] = (int)(ip[2])
        message[25] = (int)(ip[3])
        # 服务端端口（按大端输出）
        message[26:28] = server_port.to_bytes(2, 'big')
        # 校验和（按小端输出）
        checksum = crc16(message).to_bytes(2, 'little')

        return message + checksum


import socket
import json
import threading
import logging
import time
import queue
from abc import ABC, abstractmethod
from typing import Dict, List, Callable, Optional, Tuple
from loguru import logger

from rsc import UdpBroadcaster

# ----------------------
# Constants and Exceptions
# ----------------------
UDP_DISCOVERY_PORT = 50000
TCP_COMMUNICATION_PORT = 50001
PROXY_PORT = 50002
BUFFER_SIZE = 4096

class DeviceError(Exception):
    """Base exception for device related errors"""

class DeviceNotFoundError(DeviceError):
    """Requested device not found"""

class ConnectionError(DeviceError):
    """Device connection failure"""

class UnsupportedFeatureError(DeviceError):
    """Requested feature not supported"""
    
# ----------------------
# Data Structures
# ----------------------
class DeviceInfo:
    def __init__(self, serial: str, device_type: str, ip: str, tcp_port: int):
        self.serial = serial
        self.type = device_type
        self.ip = ip
        self.tcp_port = tcp_port

    def __repr__(self):
        return f"<Device {self.serial} ({self.type}) @ {self.ip}:{self.tcp_port}>"
    
    
class SecureTCPConnection:
    def __init__(self, ip: str, port: int):
        self.ip = ip
        self.port = port
        self._socket = None
        self._lock = threading.Lock()
        self._connected = False
        self._logger = logging.getLogger("Connection")
        self._message_queue = queue.Queue()
        self._receive_thread = None

    def connect(self):
        """Establish and maintain TCP connection"""
        with self._lock:
            if self._connected:
                return

            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self._socket.connect((self.ip, self.port))
                self._connected = True
                self._start_receive_thread()
            except socket.error as e:
                raise ConnectionError(f"Connection failed: {str(e)}")

    def _start_receive_thread(self):
        self._receive_thread = threading.Thread(
            target=self._receive_worker,
            daemon=True
        )
        self._receive_thread.start()

    def _receive_worker(self):
        while self._connected:
            try:
                data = self._socket.recv(BUFFER_SIZE)
                if not data:
                    break
                self._message_queue.put(data)
            except (socket.error, ConnectionResetError):
                break
        self.disconnect()

    def disconnect(self):
        """Close TCP connection"""
        with self._lock:
            if self._connected and self._socket:
                self._socket.close()
                self._connected = False

    def send(self, data: bytes):
        """Send data through TCP connection"""
        with self._lock:
            if not self._connected:
                raise ConnectionError("Not connected")
            try:
                self._socket.sendall(data)
            except socket.error as e:
                raise ConnectionError(f"Send failed: {str(e)}")

    def receive(self) -> Optional[bytes]:
        """Get received data from queue"""
        try:
            return self._message_queue.get_nowait()
        except queue.Empty:
            return None

# ----------------------
# Device Abstraction Layer
# ----------------------
class BaseDevice(ABC):
    def __init__(self, info: DeviceInfo, connection: SecureTCPConnection):
        self.info = info
        self.conn = connection
        self._data_callbacks: List[Callable[[bytes], None]] = []
        self._command_callbacks: List[Callable[[dict], None]] = []
        self._stream_active = False
        self._stream_thread: Optional[threading.Thread] = None

    @property
    @abstractmethod
    def supports_acquisition(self) -> bool:
        pass

    @property
    @abstractmethod
    def supports_stimulation(self) -> bool:
        pass

    def register_data_callback(self, callback: Callable[[bytes], None]):
        self._data_callbacks.append(callback)

    def register_command_callback(self, callback: Callable[[dict], None]):
        self._command_callbacks.append(callback)

    # 启动信号采集
    def start_acq(self, persist=False, path=None):    
        self.send_command({"type": "acquire", "param": "start", "value": persist})
        if not self.supports_acquisition:
            raise UnsupportedFeatureError("signal acquisition not supported")

        self._stream_active = True
        self._stream_thread = threading.Thread(
            target=self._stream_worker,
            daemon=True
        )
        self._stream_thread.start()
      
    # 停止信号采集  
    def stop_acq(self):    
        self._stream_active = False
        if self._stream_thread:
            self._stream_thread.join()
    
    # 设置信号采集配置
    def set_acq_config(self, config: dict): 
        """Set data acquisition configuration"""
        self.send_command({"type": "acquire", "param": "config", "value": config})

    # 设置电刺激配置
    def set_stim_config(self, config: dict):
        """Set stimulation configuration"""
        self.send_command({"type": "stim", "param": "config", "value": config})

    def start_stim(self):
        """Start stimulation"""
        self.send_command({"type": "stim", "param": "start"})

    def _stream_worker(self):
        while self._stream_active:
            data = self.conn.receive()
            if data:
                for callback in self._data_callbacks:
                    callback(data)
            time.sleep(0.001)

    def send_command(self, command: dict):
        """Send command and notify callbacks"""
        try:
            self.conn.send(json.dumps(command).encode())
            for callback in self._command_callbacks:
                callback(command)
        except ConnectionError as e:
            logging.error(f"Command failed: {str(e)}")

# ----------------------
# Device Implementations
# ----------------------
class LJS1(BaseDevice):
    @property
    def supports_acquisition(self) -> bool:
        return True

    @property
    def supports_stimulation(self) -> bool:
        return False

    def set_sampling_rate(self, rate: int):
        self.send_command({"type": "config", "param": "rate", "value": rate})

class ARS(BaseDevice):
    @property
    def supports_acquisition(self) -> bool:
        return True

    @property
    def supports_stimulation(self) -> bool:
        return False
    def set_stimulation(self, intensity: float, duration: float):
        self.send_command({
            "type": "stimulate",
            "intensity": intensity,
            "duration": duration
        })

class X8(BaseDevice):
    @property
    def supports_acquisition(self) -> bool:
        return True

    @property
    def supports_stimulation(self) -> bool:
        return False
    def set_stimulation(self, intensity: float, duration: float):
        self.send_command({
            "type": "stimulate",
            "intensity": intensity,
            "duration": duration
        })
 
class C64RS(BaseDevice):
    @property
    def supports_acquisition(self) -> bool:
        return True

    @property
    def supports_stimulation(self) -> bool:
        return False
    def set_stimulation(self, intensity: float, duration: float):
        self.send_command({
            "type": "stimulate",
            "intensity": intensity,
            "duration": duration
        })       
        
class C256RS(BaseDevice):
    @property
    def supports_acquisition(self) -> bool:
        return True

    @property
    def supports_stimulation(self) -> bool:
        return False
    def set_stimulation(self, intensity: float, duration: float):
        self.send_command({
            "type": "stimulate",
            "intensity": intensity,
            "duration": duration
        })

# ----------------------
# Proxy Server
# ----------------------
class DeviceProxy:
    def __init__(self, sdk):
        self.sdk = sdk
        self._running = False
        self._server_socket = None
        self._clients = {}
        self._logger = logging.getLogger("DeviceProxy")

    def start(self, port: int = PROXY_PORT):
        self._running = True
        self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_socket.bind(('0.0.0.0', port))
        self._server_socket.listen(5)
        threading.Thread(target=self._accept_clients, daemon=True).start()

    def stop(self):
        self._running = False
        if self._server_socket:
            self._server_socket.close()

    def _accept_clients(self):
        while self._running:
            try:
                client_socket, addr = self._server_socket.accept()
                client_id = f"{addr[0]}:{addr[1]}"
                self._clients[client_id] = client_socket
                threading.Thread(
                    target=self._handle_client,
                    args=(client_socket, addr),
                    daemon=True
                ).start()
            except OSError:
                break

    def _handle_client(self, client_socket: socket.socket, addr: Tuple[str, int]):
        client_id = f"{addr[0]}:{addr[1]}"
        self._logger.info(f"New client connected: {client_id}")

        while self._running:
            try:
                data = client_socket.recv(BUFFER_SIZE)
                if not data:
                    break

                # Process proxy command
                try:
                    command = json.loads(data.decode())
                    self._process_command(client_socket, command)
                except json.JSONDecodeError:
                    self._logger.warning(f"Invalid command from {client_id}")

            except (ConnectionResetError, socket.error):
                break

        self._logger.info(f"Client disconnected: {client_id}")
        client_socket.close()
        del self._clients[client_id]

    def _process_command(self, client_socket: socket.socket, command: dict):
        required_fields = ["device_id", "action", "params"]
        if not all(field in command for field in required_fields):
            self._logger.warning("Invalid command structure")
            return

        try:
            device = self.sdk.get_device(command["device_id"])
            if not device:
                raise DeviceNotFoundError

            # Execute device command
            if command["action"] == "send":
                device.send_command(command["params"])
                client_socket.send(b"Command executed")
            elif command["action"] == "stream":
                self._setup_streaming(client_socket, device)
            else:
                client_socket.send(b"Invalid action")

        except DeviceError as e:
            client_socket.send(f"Error: {str(e)}".encode())

    def _setup_streaming(self, client_socket: socket.socket, device: BaseDevice):
        def forward_data(data: bytes):
            try:
                client_socket.send(data)
            except (socket.error, ConnectionResetError):
                device.stop_data_stream()

        device.register_data_callback(forward_data)
        device.start_data_stream()
        client_socket.send(b"Streaming started")


from qlsdk.sdk import Hub
from qlsdk.x8 import X8
from loguru import logger


class X8M(Hub):
    def __init__(self):
        super().__init__()
        self._devices: dict[str, X8] = {}
        self._search_running = False
        self._search_timer = None
    
    def add_device(self, mac: str):
        if mac in list(self._devices.keys()):
            self._devices[mac].update_info()
            logger.debug(f"update x8 device mac: {mac}")
        else:
            dev = X8(mac)
            if dev.connected:
                self._devices[mac] = dev
                logger.info(f"add x8 device mac: {dev}")


from qlsdk.core.utils import to_channels
from loguru import logger

class Packet(object):
    def __init__(self):
        self.time_stamp = None
        self.pkg_id = None
        self.result = None
        self.channels = None
        
            
class RscPacket(Packet):
    def __init__(self):
        super().__init__()
        self.origin_sample_rate = None
        self.sample_rate = None
        self.sample_num = None
        self.resolution = None
        self.filter = None
        self.data_len = None
        self.trigger = None
        self.eeg = None # 数字信号
        self.eeg_p = None # 物理值
    
    @staticmethod
    def transfer(body: bytes) -> 'RscPacket':
        packet = RscPacket()
        packet.time_stamp = int.from_bytes(body[0:8], 'little')
        packet.result = body[8]
        packet.pkg_id = int.from_bytes(body[9: 13], 'little')
        logger.trace(f"pkg_id: {packet.pkg_id}")
        packet.channels = to_channels(body[13: 45])
        packet.origin_sample_rate = int.from_bytes(body[45: 49], 'little')
        packet.sample_rate = int.from_bytes(body[49: 53], 'little')
        packet.sample_num = int.from_bytes(body[53: 57], 'little')
        packet.resolution = int(int(body[57]) / 8)
        packet.filter = body[58]
        packet.data_len = int.from_bytes(body[59: 63], 'little')
        # 步径 相同通道的点间隔
        step = int(len(packet.channels) * packet.resolution + 4)
        b_eeg = body[63:]
        ch_num = len(packet.channels) 
        # 字节序列(4Cn{channel_size}){sample_num})
        packet.trigger = [int.from_bytes(b_eeg[j * step : j * step + 4], 'little', signed=False) for j in range(packet.sample_num)]
        
        packet.eeg = [
            [
                int.from_bytes(b_eeg[i * packet.resolution + 4 + j * step:i * packet.resolution + 4 + j * step + 3], 'big', signed=True)
                for j in range(packet.sample_num)
            ]
            for i in range(ch_num)
        ]
                
        logger.trace(packet)
        
        return packet

    def copy(self) -> 'RscPacket':
        packet = RscPacket()
        packet.time_stamp = self.time_stamp
        packet.pkg_id = self.pkg_id
        packet.result = self.result
        packet.channels = self.channels
        packet.origin_sample_rate = self.origin_sample_rate
        packet.sample_rate = self.sample_rate
        packet.sample_num = self.sample_num
        packet.resolution = self.resolution
        packet.filter = self.filter
        packet.data_len = self.data_len
        packet.trigger = self.trigger
        packet.eeg = self.eeg
        return packet  
        
    def __str__(self):
        return f"""[
            "time_stamp": {self.time_stamp}
            "pkg_id": {self.pkg_id}
            "channels": {self.channels}
            "trigger": {self.trigger}
            "eeg": {self.eeg}
        ]"""
        
class ImpedancePacket(Packet):
    def __init__(self):
        super().__init__()
        self.impedance = None
        
    @staticmethod
    def transfer(body:bytes) -> 'ImpedancePacket':
        packet = ImpedancePacket()
        packet.time_stamp = int.from_bytes(body[0:8], 'little')
        # packet.result = body[8]
        packet.pkg_id = int.from_bytes(body[9: 13], 'little')
        packet.channels = to_channels(body[13: 45])
        # packet.sample_rate = int.from_bytes(body[45: 49], 'little')
        # packet.sample_len = int.from_bytes(body[49: 53], 'little')
        # packet.resolution = int(int(body[53]) / 8)
        # packet.filter = int(int(body[54]) / 8)
        # packet.wave_type = int(int(body[55]) / 8)
        # packet.wave_freq = int.from_bytes(body[56: 60], 'little')
        # packet.data_len = int.from_bytes(body[60: 64], 'little')
        b_impedance = body[64:]
        packet.impedance = [int.from_bytes(b_impedance[j * 4 : j * 4 + 4], 'little', signed=False) for j in range(len(packet.channels))]

        logger.trace(f"impedance: {packet}")
        
        return packet

    def __str__(self):
        return f"""[
            "time_stamp": {self.time_stamp}
            "pkg_id": {self.pkg_id}
            "channels": {self.channels}
            "impedance": {self.impedance}
        ]"""
        
    
class C256RSPacket(Packet):
    def __init__(self):
        super().__init__()
        self.origin_sample_rate = None
        self.sample_rate = None
        self.sample_num = None
        self.resolution = None
        self.filter = None
        self.data_len = None
        self.trigger = None
        self.eeg = None
    
    @staticmethod
    def transfer(body: bytes) -> 'RscPacket':
        packet = RscPacket()
        packet.time_stamp = int.from_bytes(body[0:8], 'little')
        packet.result = body[8]
        packet.pkg_id = int.from_bytes(body[9: 13], 'little')
        packet.channels = to_channels(body[13: 45])
        logger.trace(f"pkg_id: {packet.pkg_id}, channels: {packet.channels}")
            
        packet.origin_sample_rate = int.from_bytes(body[45: 49], 'little')
        packet.sample_rate = int.from_bytes(body[49: 53], 'little')
        packet.sample_num = int.from_bytes(body[53: 57], 'little')
        packet.resolution = int(int(body[57]) / 8)
        packet.filter = body[58]
        packet.data_len = int.from_bytes(body[59: 63], 'little')
                
        # 数据块
        b_eeg = body[63:]
        # 根据值域分割数组-代表设备的4个模块
        ranges = [(1, 64), (65, 128), (129, 192), (193, 256)]
        sub_channels = [[x for x in packet.channels if low <= x <= high] for low, high in ranges]
        # 只处理选中的模块
        sub_channels = [_ for _ in sub_channels if len(_) > 0]
        # 步径 相同通道的点间隔
        step = int(len(packet.channels) * packet.resolution + 4 * len(sub_channels))
        offset = 0
        
        # 分按子模块处理
        for channels in sub_channels:
            logger.trace(f"子数组: {channels} 长度: {len(channels)}")
            channel_size = len(channels) 
            
            # 模块没有选中通道的，跳过
            if channel_size == 0:
                continue
            
            # 只保留第一个有效模块的trigger，其他的模块是冗余信息，无实际含义
            if packet.trigger is None:
                packet.trigger = [int.from_bytes(b_eeg[j * step : j * step + 4], 'little', signed=False) for j in range(packet.sample_num)]
                logger.trace(f"trigger: {packet.trigger}")
                trigger_positions = [index for index, value in enumerate(packet.trigger) if value != 0]
                if len(trigger_positions) > 0:
                    logger.debug(f"Trigger触发点位置: {trigger_positions}, 触发点时间戳: {[packet.time_stamp + int(pos * 1000 / packet.sample_rate) for pos in trigger_positions]}")
                
            eeg = [
                [
                    int.from_bytes(b_eeg[offset + step * j + 4 +  k * packet.resolution : offset + step * j + 7 +  k * packet.resolution], 'big', signed=True)
                    for j in range(packet.sample_num)
                ]
                for k in range(channel_size)
            ]
            packet.eeg = packet.eeg + eeg if packet.eeg else eeg
            
            offset += 4 + channel_size * packet.resolution 
            
        logger.trace(packet)
        return packet
    
  
class C256ImpedancePacket(Packet):
    def __init__(self):
        super().__init__()
        self.impedance = None
        
    @staticmethod
    def transfer(body:bytes) -> 'ImpedancePacket':
        packet = ImpedancePacket()
        packet.time_stamp = int.from_bytes(body[0:8], 'little')
        # packet.result = body[8]
        packet.pkg_id = int.from_bytes(body[9: 13], 'little')
        packet.channels = to_channels(body[13: 45])
        # packet.sample_rate = int.from_bytes(body[45: 49], 'little')
        # packet.sample_len = int.from_bytes(body[49: 53], 'little')
        # packet.resolution = int(int(body[53]) / 8)
        # packet.filter = int(int(body[54]) / 8)
        # packet.wave_type = int(int(body[55]) / 8)
        # packet.wave_freq = int.from_bytes(body[56: 60], 'little')
        # packet.data_len = int.from_bytes(body[60: 64], 'little')
        b_impedance = body[64:]
        packet.impedance = [int.from_bytes(b_impedance[j : j + 4], 'little', signed=False) for j in range(len(packet.channels))]

        logger.trace(f"impedance: {packet}")

    def __str__(self):
        return f"""
            time_stamp: {self.time_stamp}
            pkg_id: {self.pkg_id}
            result: {self.result}
            channels: {self.channels}
            data len: {self.data_len}
            impedance: {self.impedance}
        """
            


# __version__ = "0.1.1"

# 暴露公共接口
from .ar4m import AR4M
from .ar4 import AR4
from .x8 import X8
from .x8m import X8M
from .rsc import *

__all__ = ['AR4M', 'AR4', 'AR4Packet', 'X8']

# from importlib import import_module
# from pathlib import Path

# _MODULES = [p.stem for p in Path(__file__).parent.glob("*.py") if p.stem != "__init__"]

# for module in _MODULES:
#     import_module(f".{module}", package="qlsdk")


from abc import ABC, abstractmethod

class IParser(ABC):
    def __init__(self):
        super().__init__()        
        
    @abstractmethod
    def append(self, buffer):
        pass 
    
    def set_device(self, device):
        pass
    
    def start(self):
        pass
    
    def stop(self):
        pass


from multiprocessing import Lock
from qlsdk.rsc.interface import IDevice, IParser

from loguru import logger
from threading import Thread
from time import time_ns
from qlsdk.rsc.command import CommandFactory

class RSCMessageParser(IParser):
    def __init__(self, device : IDevice):
        # 待解析的数据来源于该设备
        self.device = device    
        self.running = False
        
        # 缓冲区-用于处理数据
        self.buffer:bytearray = bytearray()
        # 读写锁-用于临时缓冲区（避免读写冲突）
        self._lock = Lock()
        
    @property
    def header(self):
        return b'\x5A\xA5'
    
    @property
    def header_len(self):
        return 14
    
    @property
    def cmd_pos(self):
        return 12
    
    def set_device(self, device):
        self.device = device
        
    def append(self, value):
        print(f"append value len: {len(value)}")
    def __parser__(self):
        logger.info("数据解析开始")
                
        try:
            while self.running:
                
                # 查找消息头
                while len(self.buffer) < len(self.header):
                    read_data = self.device.read_msg(1)
                    if read_data is None:
                        if self.running:
                            continue
                        else:
                            break
                    if read_data[0] == self.header[len(self.buffer)]:
                        self.buffer.extend(read_data)
                    else:
                        self.buffer.clear()
                        break            
                
                if len(self.buffer) < len(self.header):
                    continue
                
                
                # 判断消息头
                while len(self.buffer) < self.header_len:
                    read_data = self.device.read_msg(self.header_len - len(self.buffer))
                    if read_data is None:
                        if self.running:
                            continue
                        else:
                            break
                    self.buffer.extend(read_data)                    
                
                # # 移动下标(指向包长度的位置)
                # self.buffer.seek(self.start_pos + 8)
                # 包总长度
                pkg_len = int.from_bytes(self.buffer[8:12], 'little')
                # 256*32K = 8388608
                # 太长2M
                if pkg_len > 2048000:
                    print(f"error message pkg_len={pkg_len} > 1000000")
                    self.buffer.clear()
                    continue
                # 太短
                if pkg_len < self.header_len:
                    print(f"error message pkg_len={pkg_len} < {self.header_len}")
                    self.buffer.clear()
                    continue
                
                # 读取消息体
                while len(self.buffer) < pkg_len:
                    read_data = self.device.read_msg(pkg_len - len(self.buffer))
                    if read_data is None:
                        if self.running:
                            continue
                        else:
                            break
                    self.buffer.extend(read_data)
                
                if len(self.buffer) < pkg_len:
                    continue
                    
                # msg_crc = self.buffer[-2:]
                # cal_crc = (bytes_buffer, len(bytes_buffer) - 2)
                # if cal_crc != msg_crc:
                #     print(f'crc fail {cal_crc} {msg_crc}')
                # if cal_crc != msg_crc:
                #     print(f'crc fail {cal_crc} {msg_crc}')                    
                
                self.unpack(self.buffer)
                self.buffer.clear()
                
        except Exception as e:
            logger.error(f"数据解析异常: {e}")
                
        logger.info(f"数据解析结束:{self.running}")
    
    def unpack(self, packet):        
        # 提取指令码
        cmd_code = int.from_bytes(packet[self.cmd_pos : self.cmd_pos + 2], 'little')
        cmd_class = CommandFactory.create_command(cmd_code)
        instance = cmd_class(self.device)
        instance.parse_body(packet[self.header_len:-2])
        return instance
            
    def start(self):
        self.running = True
        parser = Thread(target=self.__parser__, daemon=True)
        parser.start()
        
    def stop(self):
        self.running = False
        


from multiprocessing import Queue
from threading import Thread
from time import sleep, time_ns
from typing import Any, Dict, Literal

from loguru import logger
import numpy as np
from qlsdk.core.entity import RscPacket, ImpedancePacket
from qlsdk.core.utils import to_bytes
from qlsdk.rsc.interface import IDevice, IParser
from qlsdk.rsc.command import SetImpedanceParamCommand, StartImpedanceCommand, StopImpedanceCommand, StartStimulationCommand, StopStimulationCommand, SetAcquisitionParamCommand, StartAcquisitionCommand, StopAcquisitionCommand
from qlsdk.rsc.paradigm import StimulationParadigm
from qlsdk.rsc.parser.base import TcpMessageParser

    
def intersection_positions(A, B):
    setB = set(B)
    seen = set()
    return [idx for idx, elem in enumerate(A) 
            if elem in setB and elem not in seen and not seen.add(elem)]    
    
class QLBaseDevice(IDevice):
    
    __TRIGGER_MAPPING = {
        0x3E8: "Start of stimulation",
        0x3E9: "End of stimulation",
        0x3EA: "Ascending end of stimulation",
        0x3EB: "Descending start of stimulation ",
        0x3EC: "刺激参数有误",
        0x3ED: "End of stimulation (by force)",
    }    
    def __init__(self, socket):
        self.socket = socket
        
        # 启动数据解析线程
        self._parser: IParser = None
        
        # 用在edf/bdf文件写入中，不建议使用
        self._record_duration = None    
        
        # 设备信息
        self.device_id = None
        self.device_name = None
        self._device_type = None
        self._device_no = None
        self.software_version = None
        self.hardware_version = None
        self.connect_time = None
        self.current_time = None
        # mV
        self.voltage = None
        # %
        self.battery_remain = None
        # %
        self.battery_total = None
        # persist
        self._recording = False
        self._storage_path = None
        self._file_prefix = None
        
        # 可设置参数
        # 采集：采样量程、采样率、采样通道
        # 刺激：刺激电流、刺激频率、刺激时间、刺激通道
        # 采样量程（mV）：188、375、563、750、1125、2250、4500
        self._sample_range:Literal[188, 375, 563, 750, 1125, 2250, 4500] = 188        
        # 采样率（Hz）：250、500、1000、2000、4000、8000、16000、32000
        self._sample_rate:Literal[250, 500, 1000, 2000, 4000, 8000, 16000, 32000] = 500
        self._physical_max = self._sample_range * 1000  # 物理最大值（uV）
        self._physical_min = -self._sample_range * 1000  # 物理最小值（uV）
        self._digital_max = 8388607
        self._digital_min = -8388608
        self._physical_range = self._physical_max - self._physical_min
        self._digital_range = 16777215
        self._acq_channels = None
        self._acq_param = {
            "sample_range": 188,
            "sample_rate": 500,
            "channels": [],
        }
        
        self._stim_param = None
        
        self._impedance_channels = []
        
        self.stim_paradigm: StimulationParadigm = None
        # 信号采集状态
        # 信号数据包总数（一个信号采集周期内）
        # 信号采集参数
        # 电刺激状态
        # 电刺激开始时间（最近一次）
        # 电刺激结束时间（最近一次）
        # 电刺激参数
        # 启动数据解析线程
        # 数据存储状态
        # 存储目录
        
        # 
        self._signal_consumer: Dict[str, Queue[Any]]={}
        self._impedance_consumer: Dict[str, Queue[Any]]={}
        
        # EDF文件处理器
        self._edf_handler = None       
        self.storage_enable = True 
        self._listening = False
        # self.ready()
        self._signal_cache: Queue = None
        self._recording = False
        
        
    def parser(self) -> IParser:
        return self._parser
    
    def set_record_duration(self, record_duration: float):
        self._record_duration = record_duration      
    
    # 数据包处理
    def produce(self, body: bytes, type:Literal['signal', 'impedance']="signal"):
        if body is None: return      
        
        if type == "signal":
            self._produce_signal(body)
        elif type == "impedance":
            self._produce_impedance(body)       
        
    def _produce_impedance(self, body: bytes):
        # 分发阻抗数据包给订阅者
        if len(self._impedance_consumer) > 0:
            packet = self._impedance_wrapper(body)
            for topic, q in self._impedance_consumer.items():
                try:
                    # 队列满了就丢弃最早的数据
                    if q.full(): 
                        q.get()
                    q.put(packet, timeout=1)
                except Exception as e:
                    logger.error(f"impedance data put to queue exception: {str(e)}")
    
    def _produce_signal(self, body: bytes):        
        
        # 处理信号数据
        data = self._signal_wrapper(body)  
        # logger.debug("pkg_id: {}, eeg len: {}".format(data.pkg_id, len(data.eeg)))
        # 
        trigger_positions = [index for index, value in enumerate(data.trigger) if value != 0]
        if len(trigger_positions) > 0:
            # logger.debug(f"Trigger触发点位置: {trigger_positions}, 触发点时间戳: {[data.time_stamp + int(pos * 1000 / data.sample_rate) for pos in trigger_positions]}")
            for pos in trigger_positions:
                self.trigger(self.trigger_info(data.trigger[pos]))
        # 存储
        if self.storage_enable:            
            # 确保记录线程启动
            if self._recording is False:
                self._start_recording() 
                
            # 写入文件的缓存队列
            if self._signal_cache is None:
                self._signal_cache = Queue(256 * 1024 * 1024)  # 256MB缓存
            tmp = data.copy()
            self._signal_cache.put(tmp)
                                    
        if len(self.signal_consumers) > 0 :
            logger.trace(f"dg eeg: {data.eeg}")
            # 信号数字值转物理值
            data.eeg = self.eeg2phy(np.array(data.eeg))
            logger.trace(f"ph eeg: {data.eeg}")
            
            # 发送数据包到订阅者
            for q in list(self.signal_consumers.values()):
                # 队列满了就丢弃最早的数据
                if q.full(): 
                    q.get()
                    
                q.put(data)    
                
    def _impedance_wrapper(self, body: bytes):        
        packet = ImpedancePacket().transfer(body)    
        if self._impedance_channels is not None and len(self._impedance_channels) > 0:
            # 只保留设置的阻抗通道
            channel_pos = intersection_positions(packet.channels, self._impedance_channels)
            packet.impedance = [packet.impedance[i] for i in channel_pos]
            packet.channels = [packet.channels[i] for i in channel_pos]
            
        return packet
                
    # 信号数据转换
    def _signal_wrapper(self, body: bytes):
        return RscPacket().transfer(body)
                
    def _write_signal(self):
        # 文件写入到edf
        if self._edf_handler is None:
            logger.debug("Initializing EDF handler for data storage")
            self.init_edf_handler()
        
        while self._recording:            
            data = self._signal_cache.get()
            self._edf_handler.write(data)
            if data is None:
                break
            
        self._recording = False
    def _start_recording(self):
        if self.storage_enable is False:
            logger.trace("Storage is disabled, will not start recording")
            return  
        
        if self._signal_cache is None:
            self._signal_cache = Queue(256 * 1024 * 1024)  # 256MB缓存
            
        try: 
            self._recording = True
            t = Thread(target=self._write_signal, daemon=True)
            t.start()
            logger.info(f"开启记录")
        except Exception as e:
            logger.error(f"开启记录失败: {str(e)}")
            return
        
    def start_listening(self):        
        
        try:
            # 启动消息解析器
            self.start_message_parser()
            
            # 启动消息监听器
            self.start_message_listening()
        except Exception as e:
            logger.error(f"设备{self.device_no}准备失败: {str(e)}")
            return False
        
        return True
    
    def stop_listening(self):
        logger.trace(f"设备{self.device_no}停止socket监听")
        self._listening = False        
        self._parser.stop()
     
    def read_msg(self, size: int) -> bytes:
        try:
            self.socket.settimeout(2.0)
            return self.socket.recv(size)
        except Exception as e:
            logger.error(f"read_msg exception: {str(e)}")
            raise ValueError("read_msg exception") from e
      
    @classmethod
    def from_parent(cls, parent:IDevice) -> IDevice:
        rlt = cls(parent.socket)
        rlt.device_id = parent.device_id
        rlt._device_no = parent.device_no
        return rlt
         
    @property
    def device_type(self) -> int:
        return self._device_type
    
    def set_impedance_channels(self, channels):
        self._impedance_channels = channels
        
    def get_impedance_channels(self): 
        return self._impedance_channels    
    
    def start_message_parser(self) -> None:
        self._parser = TcpMessageParser(self)
        self._parser.start()
        logger.debug("RSC消息解析器已启动")
    
    def start_message_listening(self) -> None:
        def _accept():
            while self._listening:
                try:
                    # 缓冲区4M
                    data = self.socket.recv(4096*1024)
                    if not data:
                        logger.warning(f"设备[{self.device_name}]连接结束")
                        break
                    self._parser.append(data)
                except Exception as e:
                    logger.debug(f"设备[{self.device_name}]接收数据异常: {str(e)}")
                    break
    
        # 启动数据接收线程
        self._listening = True
        message_accept = Thread(target=_accept, daemon=True)
        message_accept.start()
        logger.debug(f"socket消息监听已启动")
        
    def set_device_type(self, type: int):
        self._device_type = type
        
    @property
    def device_no(self) -> int:
        return self._device_no
    
    def set_device_no(self, value: int):
        self._device_no = value
        
    @property
    def physical_max(self) -> int:
        return self._physical_max
    
    @property
    def physical_min(self) -> int:
        return self._physical_min
    
    @property
    def physical_range(self) -> int:
        return self._physical_range
    
    @property
    def digital_max(self) -> int:
        return self._digital_max
    
    @property
    def digital_min(self) -> int:
        return self._digital_min    
    
    @property
    def digital_range(self) -> int:
        return self._digital_range
    
    def init_edf_handler(self):
        logger.warning("init_edf_handler not implemented in base class, should be overridden in subclass")
        pass
        
    # eeg数字值转物理值
    def eeg2phy(self, digital:int):
        # 向量化计算（自动支持广播）
        return ((digital - self.digital_min) / self.digital_range) * self.physical_range + self.physical_min
    
    @property
    def edf_handler(self):
        if not self.storage_enable:
            logger.info("已关闭文件记录，不会生成edf/bdf文件")
            return None
        
        if self._edf_handler is None:
            self.init_edf_handler() 
            
        return self._edf_handler  
    
    @property
    def acq_channels(self):
        if self._acq_channels is None:
            self._acq_channels = [i for i in range(1, 63)]
        return self._acq_channels
    @property
    def sample_range(self):
        return self._sample_range if self._sample_range else 188
    @property
    def sample_rate(self):
        return self._sample_rate if self._sample_rate else 500
    @property
    def resolution(self):
        return 24
    @property
    def sample_num(self) -> int:
        return 10
    @property
    def signal_consumers(self):
        return self._signal_consumer
    
    @property
    def impedance_consumers(self):
        return self._impedance_consumer
    
    # 设置记录文件路径
    def set_storage_path(self, dir: str):
        import os
        
        abs_path = os.path.abspath(dir)
        os.makedirs(abs_path, exist_ok=True)
        self._storage_path = abs_path
       
    # 设置记录文件名称前缀 
    def set_file_prefix(self, prefix):
        self._file_prefix = prefix
       
    # socket发送数据    
    def send(self, data):
        self.socket.sendall(data)
    
    # 设置刺激参数
    def set_stim_param(self, param):
        self.stim_paradigm = param
    
    # 设置采集参数
    def set_acq_param(self, channels, sample_rate:Literal[188, 375, 563, 750, 1125, 2250, 4500] = 500, sample_range:Literal[250, 500, 1000, 2000, 4000, 8000, 16000, 32000] = 188):
        self._acq_param["channels"] = channels 
        self._acq_param["sample_rate"] = sample_rate
        self._acq_param["sample_range"] = sample_range
        self._acq_channels = channels
        self._sample_rate = sample_rate
        self._sample_range = sample_range
        
    def start_impedance(self):
        logger.info(f"[设备-{self.device_no}]启动阻抗测量")  
        # 设置数据采集参数
        # set_param_msg = SetImpedanceParamCommand.build(self).pack()
        device_id = bytes.fromhex(self.device_id)[::-1].hex() if self.device_id else '00000000'
        set_param_msg = bytes.fromhex(f'5aa50239{device_id}3f0000001104ffffffffffffffff000000000000000000000000000000000000000000000000e8030000fa00000010000164000000745c5aa50239390045243a00000012040000000000000000000000000000000000000000000000000000000000000000000001000000000000004c2a')
        logger.debug(f"set_param_msg message is {set_param_msg.hex()}")
        self.socket.sendall(set_param_msg)
        sleep(0.5)
    
        impedance_start_msg = StartImpedanceCommand.build(self).pack()
        logger.debug(f"start_impedance message is {impedance_start_msg.hex()}")
        self.socket.sendall(impedance_start_msg)
    
    def stop_impedance(self):
        logger.info(f"[设备{self.device_no}]停止阻抗测量")  
        msg = StopImpedanceCommand.build(self).pack()
        logger.trace(f"stop_impedance message is {msg.hex()}")
        self.socket.sendall(msg)
    
    def start_stimulation(self):
        if self.stim_paradigm is None:
            logger.warning("刺激参数未设置，请先设置刺激参数")
            return
        logger.info(f"[设备-{self.device_no}]启动电刺激")  
        msg = StartStimulationCommand.build(self).pack()
        logger.trace(f"start_stimulation message is {msg.hex()}")
        self.socket.sendall(msg)
        # t = Thread(target=self._stop_stimulation_trigger, args=(self.stim_paradigm.duration,), daemon=True)
        # t.start()
        
    def get_stim_param(self) -> bytes:
        return self.stim_paradigm.to_bytes()
        
    def _stop_stimulation_trigger(self, duration):
        delay = duration
        while delay > 0:
            sleep(1)
            delay -= 1
        logger.debug(f"_stop_stimulation_trigger duration: {duration}")
        if self.edf_handler:
            self.edf_handler.trigger("stimulation should be stopped")
        else:
            logger.warning("stop stim trigger fail. no edf writer alive")
        
    def stop_stimulation(self):
        logger.info(f"[设备-{self.device_no}]停止电刺激")  
        msg = StopStimulationCommand.build(self).pack()
        logger.trace(f"stop_stimulation message is {msg.hex()}")
        self.socket.sendall(msg)
        
    # 启动采集
    def start_acquisition(self):
        logger.info(f"[设备-{self.device_no}]启动信号采集")  
        # 记录准备
        self._start_recording()
        # 设置数据采集参数
        param_bytes = SetAcquisitionParamCommand.build(self).pack()
        # 启动数据采集
        start_bytes = StartAcquisitionCommand.build(self).pack()
        msg = param_bytes + start_bytes
        logger.trace(f"start_acquisition message is {msg.hex()}")
        self.socket.sendall(msg)
        
    # 停止采集
    def stop_acquisition(self):
        logger.info(f"[设备-{self.device_no}]停止信号采集")  
        msg = StopAcquisitionCommand.build(self).pack()
        logger.trace(f"stop_acquisition message is {msg.hex()}")
        self.socket.sendall(msg)
        # 结束标识
        if self._signal_cache:
            # 发送结束标识
            self._signal_cache.put(None)
        
    '''
        订阅数据
        topic: 订阅主题
        q: 数据队列
        type: 数据类型，signal-信号数据，impedance-阻抗数据
    '''
    def subscribe(self, topic:str=None, q : Queue=None, type : Literal["signal","impedance"]="signal"):  
            
        # 队列名称     
        if topic is None:
            topic = f"{self.device_no}_{type}_{time_ns()}"
            
        logger.debug(f"[设备-{self.device_no}]订阅数据流： {topic}, type: {type}")    
            
        # 数据队列
        if q is None:
            q = Queue(maxsize=100 * 1024 * 1024)
        
        # 订阅生理电信号数据
        if type == "signal":
            # topic唯一，用来区分不同的订阅队列（下同）
            if topic in list(self._signal_consumer.keys()):
                logger.warning(f"已存在主题[{topic}]的信号数据订阅！")
            else:
                self._signal_consumer[topic] = q
            
        # 订阅阻抗数据
        if type == "impedance":
            if topic in list(self._impedance_consumer.keys()):
                logger.warning(f"已存在主题[{topic}]的阻抗数据订阅！")
            else:
                self._impedance_consumer[topic] = q
            
        return topic, q
    
    def trigger(self, desc):
        if self.edf_handler:
            self.edf_handler.trigger(desc)
        else:
            logger.info("已关闭文件记录，不会记录trigger信息")
        
    # 设置信号采集参数
    def gen_set_acquirement_param(self) -> bytes:
        
        body = to_bytes(self.acq_channels) 
        body += self.sample_range.to_bytes(4, byteorder='little')
        body += self.sample_rate.to_bytes(4, byteorder='little')
        body += self.sample_num.to_bytes(4, byteorder='little')
        body += self.resolution.to_bytes(1, byteorder='little')
        body += bytes.fromhex('00')     
        
        return body
    # 设置阻抗测量参数
    def gen_set_impedance_param(self) -> bytes:
        
        # 仅通道生效 32字节，其他不生效-272字节，实际73字节
        body = to_bytes(self._impedance_channels) 
        # 100 bytes
        # body += bytes.fromhex('00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000')     
        # # 100 bytes
        # body += bytes.fromhex('00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000')     
        # 73 bytes
        body += bytes.fromhex('00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000')     
        
        return bytes.fromhex('ffffffffffffffff000000000000000000000000000000000000000000000000e8030000fa00000010000164000000745c5aa50239390045243a0000001204000000000000000000000000000000000000000000000000000000000000000000000100000000000000')
    def disconnect(self):
        logger.info(f"[断开设备-{self.device_no}]的连接...")  
        self._listening = False
        try:
            sleep(0.1)
            self.socket.shutdown(2)
            self.socket.close()
            logger.info(f"[设备-{self.device_no}]设备连接已断开")  
        except Exception as e:
            logger.error(f"断开设备连接异常: {str(e)}")
         
        # 关闭解析器
        self._parser.stop()   
        
    def enable_storage(self, enable: bool = True):
        self.storage_enable = enable
        
    def trigger_info(self, code: int) -> str:
        return QLBaseDevice.__TRIGGER_MAPPING.get(code, hex(code))    
    
    def __str__(self):
        return f''' 
            Device: 
                Name: {self.device_name}, 
                Type: {hex(self.device_type) if self.device_type else None}, 
                ID: {self.device_no if self.device_no else None}, 
                Software: {self.software_version}, 
                Hardware: {self.hardware_version}, 
                Connect Time: {self.connect_time},
                Current Time: {self.current_time},
                Voltage: {str(self.voltage) + "mV" if self.voltage else None},
                Battery Remain: {str(self.battery_remain)+ "%" if self.battery_remain else None},
                Battery Total: {str(self.battery_total) + "%" if self.battery_total else None}
            '''

    def __eq__(self, other:IDevice):
        return self.device_type == other.device_type and self.device_no == other.device_no

    def __hash__(self):
        return hash((self.device_type, self.device_no))    


from loguru import logger
from qlsdk.core.entity import C256RSPacket
from qlsdk.persist import RscEDFHandler
from qlsdk.rsc.device.device_factory import DeviceFactory
from qlsdk.rsc.command import *
from qlsdk.rsc.device.base import QLBaseDevice

class C256RS(QLBaseDevice):
    # C256RS设备类型标识符
    device_type_001 = 0x51  
    device_type = 0x45  
    
    def __init__(self, socket):
        super().__init__(socket)
        
    @classmethod
    def from_parent(cls, parent:IDevice) -> IDevice:
        rlt = cls(parent.socket)
        rlt.device_id = parent.device_id
        rlt._device_no = parent.device_no
        return rlt
        
    def init_edf_handler(self):
        self._edf_handler = RscEDFHandler(self.sample_rate,  self.sample_range * 1000 , - self.sample_range * 1000, self.resolution)  
        self._edf_handler.set_device_type(self.device_type)
        self._edf_handler.set_device_no(self.device_no)
        self._edf_handler.set_storage_path(self._storage_path)
        self._edf_handler.set_file_prefix(self._file_prefix if self._file_prefix else 'C256RS')
        logger.debug(f"EDF Handler initialized")
      
    @property
    def acq_channels(self):
        if self._acq_channels is None:
            self._acq_channels = [i for i in range(1, 256)]
        return self._acq_channels  
    
    def _signal_wrapper(self, body: bytes):
        return C256RSPacket().transfer(body)
        
    def get_stim_param(self) -> bytes:
        return self.stim_paradigm.to_bytes_c256()    
    
DeviceFactory.register(C256RS.device_type, C256RS) 
DeviceFactory.register(C256RS.device_type_001, C256RS) 


from loguru import logger
from threading import Thread, Lock
import socket

class DeviceProxy(object):
    def __init__(self, client_socket):
        self.client_socket = client_socket
        self.server_socket = None   
        # 客户端设备
        # self.device = QLDevice(client_socket)
        # self.parser = DeviceParser(self.device)
        
        self._init_server()
        
    def _init_server(self):
        # 连接目标服务（桌面端运行在本地的18125端口）
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.connect(('localhost', 18125))
    
    def start(self):
        logger.info("DeviceProxy连接已建立")
        
        try:            
            # 创建双向转发线程
            client_thread = Thread(
                target=self.data_forward,
                args=(self.client_socket, self.server_socket, "设备->软件")
            )
            server_thread = Thread(
                target=self.data_forward,
                args=(self.server_socket, self.client_socket, "软件->设备")
            )

            # 启动线程
            client_thread.start()
            server_thread.start()

            # 等待线程结束
            client_thread.join()
            server_thread.join()
            
        except Exception as e:
            logger.error(f"连接异常: {str(e)}")
            self.client_socket.close()      
        finally:
            logger.error(f"连接已结束")
            self.client_socket.close()
            
    def data_forward(self, src, dest, direction, callback=None):
        """
        数据转发函数
        :param src: 源socket连接
        :param dest: 目标socket连接
        :param direction: 转发方向描述
        """
        try:
            logger.debug(f"[{direction}] 转发开始")
            while True:
                data = src.recv(4096*1024)  # 接收数据缓冲区设为4KB
                if not data:
                    break
                logger.info(f"[{direction}] 转发指令: {hex(int.from_bytes(data[12:14], 'little'))} ")
                logger.debug(f"[{direction}] 转发 {data.hex()} ")
                if callback:
                    callback(data)
                dest.sendall(data)
                logger.debug(f"[{direction}] 转发 {len(data)} 字节")
        except ConnectionResetError:
            logger.error(f"[{direction}] 连接已关闭")
        finally:
            src.close()
            dest.close()
            
    # def device_message(self, data):     
    #     # 解析数据包
    #     self.parser.append(data)


from enum import Enum
import threading
import collections
from typing import Optional

from loguru import logger
from qlsdk.rsc.interface import IDevice, IParser
from qlsdk.rsc.command import CommandFactory

# 解析状态机
class ParserState(Enum):
    FIND_HEADER = 1
    READ_LENGTH = 2
    READ_BODY = 3

class TcpMessageParser(IParser):
    HEADER = b'\x5a\xa5'
    HEADER_LEN = 14
    MAX_PKG_LEN = 1 * 1024 * 1024        # 1 MB
    MAX_BUF_LEN = 100 * 1024 * 1024       # 10 MB
    READ_CHUNK = 4096

    def __init__(self, device: IDevice):
        self.device = device

        # 生产者-消费者缓冲区：线程安全 deque + bytearray
        self._buf = bytearray()
        self._lock = threading.Lock()
        self._not_empty = threading.Condition(self._lock)

        self._running = threading.Event()

    # ---------- 生产者 ----------
    def append(self, data: bytes) -> None:
        logger.info(f"接收数据: {data.hex()}")
        with self._not_empty:
            if len(self._buf) + len(data) > self.MAX_BUF_LEN:
                logger.warning("缓冲区超限，丢弃旧数据")
                self._buf.clear()
            self._buf.extend(data)
            self._not_empty.notify()

    # ---------- 消费者 ----------
    def start(self) -> None:
        self._running.set()
        self._thread = threading.Thread(target=self._parser_loop, daemon=True)
        self._thread.start()

    def stop(self) -> None:
        self._running.clear()
        with self._not_empty:
            self._not_empty.notify_all()
        self._thread.join(timeout=1)

    # ---------- 解析主循环 ----------
    def _parser_loop(self) -> None:
        """状态机：找头 -> 读长度 -> 读体 -> 校验 -> 投递"""
        state = ParserState.FIND_HEADER
        need = len(self.HEADER)
        header_pos = 0
        pkg_len = 0
        logger.info("数据解析开始")

        while self._running.is_set():
            with self._not_empty:
                while len(self._buf) < need:
                    if not self._running.is_set():
                        return
                    self._not_empty.wait(timeout=0.1)

                view = memoryview(self._buf)  # 零拷贝视图

                if state == ParserState.FIND_HEADER:
                    # idx = view.find(self.HEADER)
                    # if idx == -1:
                    #     # 整段数据都没有头，全部丢弃
                    #     del self._buf[:len(self._buf) - 1]
                    #     continue
                    # # 去掉头部之前可能残留的脏数据
                    # del self._buf[:idx]
                    # state = 'READ_LENGTH'
                    # need = self.HEADER_LEN
                    # continue
                    with self._not_empty:
                        while len(view) < need:
                            if not self._running.is_set():
                                return
                            self._not_empty.wait(timeout=0.1)

                    # 1. 用 bytearray.find 找包头
                    idx = view.find(self.HEADER)
                    if idx == -1:
                        # 没有包头，保留最后 len(self.HEADER)-1 个字节即可
                        del self._buf[:-len(self.HEADER) + 1 or None]
                        continue

                    # 2. 去掉头部的脏数据
                    del self._buf[:idx]
                    state = 'READ_LENGTH'
                    need = self.HEADER_LEN
                    continue

                if state == 'READ_LENGTH':
                    pkg_len = int.from_bytes(view[8:12], 'little')
                    if pkg_len < self.HEADER_LEN or pkg_len > self.MAX_PKG_LEN:
                        logger.warning(f"非法包长度 {pkg_len}，丢弃")
                        del self._buf[:1]        # 跳过当前头继续找下一个
                        state = 'FIND_HEADER'
                        need = len(self.HEADER)
                        continue

                    state = 'READ_BODY'
                    need = pkg_len
                    continue

                if state == 'READ_BODY':
                    packet = bytes(view[:pkg_len])      # 拷贝一份完整包
                    del self._buf[:pkg_len]             # 从缓冲区删除
                    self._dispatch(packet)
                    state = 'FIND_HEADER'
                    need = len(self.HEADER)

    # ---------- 业务分发 ----------
    def _dispatch(self, packet: bytes) -> None:
        try:
            cmd = int.from_bytes(packet[12:14], 'little')
            cls = CommandFactory.create_command(cmd)
            inst = cls(self.device)
            inst.parse_body(packet[self.HEADER_LEN:-2])  # 去掉头尾
        except Exception as exc:
            logger.exception(f"解析失败: {exc}")

    # ---------- 工具 ----------
    def set_device(self, device: IDevice) -> None:
        self.device = device


import psutil
import socket

def get_active_interfaces():
    interfaces = {}
    # 获取所有接口地址信息
    all_addrs = psutil.net_if_addrs()
    # 获取接口状态（是否处于UP状态）
    stats = psutil.net_if_stats()
    
    for name, addrs in all_addrs.items():
        # 检查接口是否启用
        if stats[name].isup:
            ips = []
            for addr in addrs:
                # 提取IPv4和IPv6地址
                if addr.family == socket.AF_INET:
                    ips.append(f"IPv4: {addr.address}")
                elif addr.family == socket.AF_INET6:
                    ips.append(f"IPv6: {addr.address}")
            # 过滤无IP的接口（可选）
            if ips:
                interfaces[name] = {
                    "status": "UP",
                    "IPs": ips
                }
    return interfaces

# 调用并打印结果
active_ifs = get_active_interfaces()
for iface, info in active_ifs.items():
    print(f"接口: {iface}")
    print(f"状态: {info['status']}")
    print(f"IP地址: {info['IPs']}\n")


from time import sleep
from loguru import logger
import os
from threading import Thread

from qlsdk import AR4M

#------------------------------------------------------------------
# 日志文件配置
#------------------------------------------------------------------
LOG_DIR = os.path.expanduser("./logs")
LOG_FILE = os.path.join(LOG_DIR, "app_{time}.log")
if not os.path.exists(LOG_DIR):
    os.mkdir(LOG_DIR)

logger.add(LOG_FILE, rotation = "50MB")     

def consumer(q):
    t = Thread(target=deal_data, args=(q,))
    t.daemon = True
    t.start()
        
def deal_data(q):
    while True:
        data = q.get()
        if data is None:
            break
        logger.info(data)

# 主函数
if __name__ == "__main__":
    try:
        ar4m = AR4M()
        ar4m.search()
        sleep(6)
        for dev in list( ar4m.devices.values()):
            ret = dev.start_acquisition()
            topic, queue = dev.subscribe()
            logger.info(f"启动{dev.box_mac}的数据采集{'成功' if ret else '失败'}")
            
        sleep(60)
        
        for dev in list( ar4m.devices.values()):
            ret = dev.stop_acquisition()
            dev.get_acq_start_time() 
            logger.info(f"关闭{dev.box_mac}的数据采集{'成功' if ret else '失败'}")
        sleep(1)
    except Exception as e:
        logger.error(f"程序运行异常: {str(e)}")
    finally:        
        logger.info("程序结束。")


from multiprocessing import Queue
from typing import Literal
from loguru import logger
from qlsdk.sdk import LMDevice, LMPacket
import numpy as np
from qlsdk.persist import EdfHandler
import time
    
# 设备对象
class AR4(LMDevice):
    def __init__(self, box_mac:str, is_persist:bool=True, storage_path:str=None):
        # 是否持久化-保存为文件
        self._is_persist = is_persist
        self._storage_path = storage_path
        self._edf_handler = None
        
        self._recording = False
        self._record_start_time = None
        
        self._acq_info = {}
        # 订阅者列表，数值为数字信号值
        self._dig_subscriber: dict[str, Queue] = {}
        # 订阅者列表，数值为物理信号值
        self._phy_subscriber: dict[str, Queue] = {}
        
        super().__init__(box_mac)
    
    def set_storage_path(self, storage_path):
        self._storage_path = storage_path

    @property
    def device_type(self):
        return "AR4"
    
       
    def eeg_accept(self, packet):        
        if len(self._dig_subscriber) > 0 or len(self._phy_subscriber) > 0:            
            for consumer in self._dig_subscriber.values():
                consumer.put(packet)
             
            if len(self._phy_subscriber) > 0:
                logger.debug(f"dig data eeg: {packet.eeg}")
                logger.debug(f"dig data acc: {packet.acc}")
                packet.eeg = self.eeg2phy(np.array(packet.eeg))
                packet.acc = self.acc2phy(np.array(packet.acc))
                logger.debug(f"phy data eeg: {packet.eeg}")
                logger.debug(f"phy data acc: {packet.acc}")
                for consumer2 in self._phy_subscriber.values():
                    consumer2.put(packet)
            
            if self._is_persist:
                if self._edf_handler is None:
                    self.start_record()
                    self._recording = True
                    self._record_start_time = packet.time_stamp
                    logger.info(f"开始记录数据: {self.box_mac}")
            
                # 处理数据包
                self._edf_handler.append(packet)
    
    def start_record(self):
        if self._is_persist:
            if self._edf_handler is None:
                self._edf_handler = EdfHandler(self.sample_frequency, self.eeg_phy_max, self.eeg_phy_min, self.eeg_dig_max, self.eeg_dig_min, storage_path=self._storage_path)
                self._edf_handler.set_device_type(self.device_type)
                self._edf_handler.set_storage_path(self._storage_path)
                
    def stop_record(self):
        if self._edf_handler:
            # 等待设备端数据传输完成
            time.sleep(0.5)
            # 添加结束标识
            self._edf_handler.append(None)
            self._edf_handler = None
            self._recording = False
            logger.info(f"停止记录数据: {self.box_mac}")
            
    def trigger(self, desc):
        if self._edf_handler:
            self._edf_handler.trigger(desc)
            
    # 订阅推送消息 
    def subscribe(self, topic: str = None, q: Queue = Queue(), value_type: Literal['phy', 'dig'] = 'phy') -> tuple[str, Queue]:
        if topic is None:
            topic = f"msg_{self.box_mac}"
        
        if value_type == 'dig':
            if topic in list(self._dig_subscriber.keys()):  
                logger.warning(f"ar4 {self.box_mac} 订阅主题已存在: {topic}")
                return topic, self._dig_subscriber[topic]
            self._dig_subscriber[topic] = q
        else:
            if topic in list(self._phy_subscriber.keys()):  
                logger.warning(f"ar4 {self.box_mac} 订阅主题已存在: {topic}")
                return topic, self._phy_subscriber[topic]
            self._phy_subscriber[topic] = q
        
        return topic, q    
    
    def eeg2phy(self, digtal):
        # 向量化计算（自动支持广播）
        return super().eeg2phy(digtal)
    
    def acc2phy(self, digtal):
        return super().eeg2phy(digtal)
            
class X8Packet(LMPacket):
    def __init__(self, data: bytes):
        super().__init__(data)
        self._data = data
        self._head = None
        self._body = None
        
    @property
    def head(self):
        return self._head
    
    @property
    def body(self):
        return self._body
    
    def parse(self):
        # 解析数据包头部和数据体
        if len(self._data) < 4:
            logger.error(f"数据包长度不足: {len(self._data)}")
            return False
        
        # 解析头部和数据体
        self._head = self._data[:4]
        self._body = self._data[4:]
        
        return True


from .base import QLBaseDevice
from .device_factory import DeviceFactory

# import devices
from .c64_rs import C64RS
from .c16_rs import C16RS
from .arskindling import ARSKindling
from .c256_rs import C256RS


class IParser:
    pass

class Parser(IParser):
    def __init__(self):
        super().__init__()
        
class MessageParser(Parser):
    def __init__(self):
        super().__init__()
        
    def parse(self, data):
        pass


import ctypes
from multiprocessing import Queue
import platform
from ctypes import (
    c_int, c_int32, c_int64, c_uint32, c_uint64, 
    c_char, c_char_p, c_void_p, c_double,
    Structure, POINTER, CFUNCTYPE
)
from typing import Literal
from loguru import logger
from time import sleep, time
import os 
import numpy as np
# from .persist import EdfHandler

real_path = os.path.realpath(__file__)
dll_path = f'{os.path.dirname(real_path)}/libs/libAr4SDK.dll'

# 加载 DLL
if platform.system() == 'Windows':
    _dll = ctypes.CDLL(dll_path)
else:
    raise NotImplementedError(f"不支持非Windows平台：{platform.system()}")

#------------------------------------------------------------------
# 基础类型定义
#------------------------------------------------------------------
class Ar4MacStr(Structure):
    _fields_ = [("str", c_char * 17)]

class Ar4Name(Structure):
    _fields_ = [("str", c_char * 21)]

class Ar4Device(Structure):
    _fields_ = [
        ("slot", c_int32),
        ("mac", c_uint64),
        ("hub_name", Ar4Name)
    ]

class Ar4NotifyData(Structure):
    _fields_ = [
        ("time_stamp", c_int64),
        ("pkg_id", c_int64),
        ("notify_id", c_int64),
        ("eeg", POINTER(c_int32)),
        ("eeg_ch_count", c_int32),
        ("eeg_count", c_int32),
        ("acc", POINTER(c_int32)),
        ("acc_ch_count", c_int32),
        ("acc_count", c_int32)
    ]

#------------------------------------------------------------------
# 回调函数类型
#------------------------------------------------------------------
FuncAr4DataNotify = CFUNCTYPE(None, c_void_p, POINTER(Ar4NotifyData))
FuncAr4TriggerNotify = CFUNCTYPE(None, c_void_p, c_int64, c_int32, c_uint32)
FuncAr4RecorderDisconnected = CFUNCTYPE(None, c_void_p)
FuncAr4RecorderConnected = CFUNCTYPE(None, c_void_p)
FuncAr4RecorderTimeout = CFUNCTYPE(None, c_void_p, c_int64)
FuncAr4RecorderPkgLost = CFUNCTYPE(c_int, c_void_p, c_int64, c_int64, c_int64)
FuncAr4GetTime = CFUNCTYPE(c_int64)
FuncAr4SDKPrint = CFUNCTYPE(None, c_char_p, ctypes.c_size_t)

#------------------------------------------------------------------
# SDK 函数定义
#------------------------------------------------------------------
# 设备枚举
_dll.ar4_sdk_enum_device.restype = POINTER(Ar4Device)
_dll.ar4_sdk_enum_hub.restype = POINTER(Ar4Name)
_dll.ar4_sdk_enum_in_hub.argtypes = [Ar4Name]

# MAC地址转换
_dll.ar4_mac_to_str.argtypes = [c_uint64]
_dll.ar4_mac_to_str.restype = Ar4MacStr
_dll.ar4_str_to_mac.argtypes = [Ar4MacStr]
_dll.ar4_str_to_mac.restype = c_uint64

# 连接管理
_dll.ar4_sdk_connect.argtypes = [c_uint64, c_void_p]
_dll.ar4_sdk_connect.restype = c_void_p
_dll.ar4_sdk_disconnect.argtypes = [c_void_p]

# AR4信息查询
_dll.ar4_sdk_get_box_name.argtypes = [c_void_p]
_dll.ar4_sdk_get_box_name.restype = c_uint64

# 数据采集控制
_dll.ar4_sdk_start_acq.argtypes = [c_void_p]
_dll.ar4_sdk_start_acq.restype = c_int
_dll.ar4_sdk_stop_acq.argtypes = [c_void_p]
_dll.ar4_sdk_stop_acq.restype = c_int
_dll.ar4_sdk_get_record_sample_rate.argtypes = [c_void_p]
_dll.ar4_sdk_get_record_sample_rate.restype = c_double
_dll.ar4_sdk_get_acq_start_time.argtypes = [c_void_p]
_dll.ar4_sdk_get_acq_start_time.restype = c_int64

# eeg信号物理量转换
_dll.ar4_sdk_get_eeg_phy_max.argtypes = [c_void_p]
_dll.ar4_sdk_get_eeg_phy_max.restype = c_double
_dll.ar4_sdk_get_eeg_phy_min.argtypes = [c_void_p]
_dll.ar4_sdk_get_eeg_phy_min.restype = c_double
_dll.ar4_sdk_get_eeg_digtal_max.argtypes = [c_void_p]
_dll.ar4_sdk_get_eeg_digtal_max.restype = c_int32
_dll.ar4_sdk_get_eeg_digtal_min.argtypes = [c_void_p]
_dll.ar4_sdk_get_eeg_digtal_min.restype = c_int32
_dll.ar4_sdk_get_eeg_phy_unit.argtypes = [c_void_p]
_dll.ar4_sdk_get_eeg_phy_unit.restype = c_char_p

# acc信号物理量转换
_dll.ar4_sdk_get_acc_phy_max.argtypes = [c_void_p]
_dll.ar4_sdk_get_acc_phy_max.restype = c_double
_dll.ar4_sdk_get_acc_phy_min.argtypes = [c_void_p]
_dll.ar4_sdk_get_acc_phy_min.restype = c_double
_dll.ar4_sdk_get_acc_digtal_max.argtypes = [c_void_p]
_dll.ar4_sdk_get_acc_digtal_max.restype = c_int32
_dll.ar4_sdk_get_acc_digtal_min.argtypes = [c_void_p]
_dll.ar4_sdk_get_acc_digtal_min.restype = c_int32
_dll.ar4_sdk_get_acc_phy_unit.argtypes = [c_void_p]
_dll.ar4_sdk_get_acc_phy_unit.restype = c_char_p

# 回调注册
_dll.ar4_sdk_register_data_notify.argtypes = [c_void_p, FuncAr4DataNotify]
_dll.ar4_sdk_register_data_notify.restype = None
_dll.ar4_sdk_register_trigger_notify.argtypes = [c_void_p, c_void_p]
_dll.ar4_sdk_register_conn_notify.argtypes = [c_void_p, FuncAr4RecorderConnected, FuncAr4RecorderDisconnected]
_dll.ar4_sdk_register_record_state_notify.argtypes = [c_void_p, FuncAr4RecorderConnected, FuncAr4RecorderDisconnected]
_dll.ar4_sdk_register_box_conn_notify.argtypes = [c_void_p, FuncAr4RecorderConnected, FuncAr4RecorderDisconnected]

# _dll对外封装类
class AR4SDK:
    @classmethod
    def enum_devices(cls):
        """枚举可用设备"""
        devices = []
        ptr = _dll.ar4_sdk_enum_device()
        if not ptr:
            logger.info("没有找到ar4设备")
            return devices
        
        index = 0
        while True:
            device = ptr[index]
            if not device or device.mac == 0:
                break
            # logger.debug(f"get device: {device.slot}, mac: {hex(device.mac)}, hub_name: {device.hub_name.str}")
            devices.append(device)
            index += 1
            
        return devices
    
    @classmethod
    def enum_hubs(cls):
        """枚举可用设备"""
        devices = []
        ptr = _dll.ar4_sdk_enum_hub()
        sleep(5)
        logger.debug(f"enum_hubs: {ptr}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


# 枚举后才能连接上
AR4SDK.enum_devices()
# 读取系统当前时间(ms)
def _get_time():
    cur_time = int(round(time()) * 1000)
    return cur_time  
       
class Packet(object):
    pass
            
class LMPacket(Packet):
    def __init__(self):
        self.time_stamp = None
        self.pkg_id = None
        self.notify_id = None
        self.eeg_ch_count = None
        self.eeg_count = None
        self.eeg = None
        self.acc_ch_count = None
        self.acc_count = None
        self.acc = None
    
    def transfer(self, data: Ar4NotifyData):
        self.time_stamp = data.time_stamp
        self.pkg_id = data.pkg_id
        self.notify_id = data.notify_id
        self.eeg_ch_count = data.eeg_ch_count
        self.eeg_count = data.eeg_count
        self.acc_ch_count = data.acc_ch_count
        self.acc_count = data.acc_count
        # 读eeg数据
        if self.eeg_ch_count and self.eeg_count:
            self.eeg = [data.eeg[i:self.eeg_count*self.eeg_ch_count:self.eeg_ch_count] for i in range(self.eeg_ch_count)]
        # 读acc数据
        if self.acc_ch_count and self.acc_count:
            self.acc = [[] for _ in range(self.acc_ch_count)]
            for i in range(self.acc_ch_count):
                self.acc[i] = [data.acc[j + (i * self.acc_count)] for j in range(self.acc_count)]
        
        return self
        
    def __str__(self):
        return f"""
            time_stamp: {self.time_stamp}
            pkg_id: {self.pkg_id}
            notify_id: {self.notify_id}
            eeg_ch_count: {self.eeg_ch_count}
            eeg_count: {self.eeg_count}
            acc_ch_count: {self.acc_ch_count}
            acc_count: {self.acc_count}
            eeg: {self.eeg}
            acc: {self.acc}
        """

class LMDevice(object):
    def __init__(self, box_mac:str):
        # 设备句柄
        self._handle = None
        # 设备基本信息
        self._box_type = None
        self._box_mac = box_mac
        self._box_id = None
        self._box_soc = None
        self._box_name = None
        self._box_version = None
        self._head_type = None
        self._head_mac = None
        self._head_version = None
        self._head_conn_state = None
        self._head_soc = None
        self._net_state = None
        # self._hub_name = hub_name
        # self._slot = slot
        self._connected = False
        self._conn_time = None
        self._last_time = None
        
        self._recording = False
        self._record_start_time = None
        
        ## eeg 数据
        self._eeg_phy_max = None
        self._eeg_phy_min = None
        self._eeg_dig_max = None
        self._eeg_dig_min = None
        self._eeg_phy_unit = 'uV'
        self._eeg_phy_range = None
        self._eeg_dig_range = None
        ## acc 数据
        self._acc_phy_max = None    
        self._acc_phy_min = None
        self._acc_dig_max = None
        self._acc_dig_min = None
        self._acc_phy_unit = 'mG'
        self._acc_phy_range = None
        self._acc_dig_range = None
        
        self._sample_frequency = 500
        
        self._acq_info = {}
        
        # 回调函数
        self._data_callback = FuncAr4DataNotify(self._wrap_data_accept())        
        self._trigger_callback = FuncAr4TriggerNotify(self._wrap_trigger_accept())
        self._connected_notify_callback = FuncAr4RecorderConnected(self._wrap_connected_notify())
        self._disconnected_notify_callback = FuncAr4RecorderDisconnected(self._wrap_disconnected_notify())
        self._box_connected_notify_callback = FuncAr4RecorderConnected(self._wrap_box_connected_notify())
        self._box_disconnected_notify_callback = FuncAr4RecorderDisconnected(self._wrap_box_disconnected_notify())
        self._start_record_notify_callback = FuncAr4RecorderConnected(self._wrap_start_record_notify())
        self._stop_record_notify_callback = FuncAr4RecorderDisconnected(self._wrap_stop_record_notify())
        
        self.init()
        
    @property
    def connected(self):
        return self._connected
    
    @property
    def sample_frequency(self):
        return self._sample_frequency
    
    @property
    def box_mac(self):
        return self._box_mac
    @property
    def slot(self):
        return self._slot
    @property
    def hub_name(self):
        return self._hub_name
    
    @property
    def eeg_phy_max(self):
        return self._eeg_phy_max
    
    @property
    def eeg_phy_min(self):
        return self._eeg_phy_min
    
    @property
    def eeg_dig_max(self):
        return self._eeg_dig_max
    
    @property
    def eeg_dig_min(self):  
        return self._eeg_dig_min
    
    @property
    def acc_phy_max(self):
        return self._acc_phy_max
    
    @property
    def acc_phy_min(self):  
        return self._acc_phy_min
    
    @property
    def acc_dig_max(self):
        return self._acc_dig_max
    
    @property
    def acc_dig_min(self):  
        return self._acc_dig_min
    
    @property
    def eeg_phy_unit(self):
        return self._eeg_phy_unit
    
    @property
    def acc_phy_unit(self):
        return self._acc_phy_unit
    
    @property
    def eeg_phy_range(self):
        return self._eeg_phy_range
    
    @property
    def eeg_dig_range(self):
        return self._eeg_dig_range
    
    @property
    def acc_phy_range(self):
        return self._acc_phy_range
    
    @property
    def acc_dig_range(self):
        return self._acc_dig_range
    
    def init(self):        
        if self.connect():
            self.get_sample_rate()
            ## eeg 参数
            self.get_eeg_phy_max()
            self.get_eeg_phy_min()
            self.get_eeg_phy_unit()
            self.get_eeg_digital_max()
            self.get_eeg_digital_min()
            self._eeg_phy_range = self._eeg_phy_max - self._eeg_phy_min
            self._eeg_dig_range = self._eeg_dig_max - self._eeg_dig_min
            
            ## acc 参数
            self.get_acc_phy_max()
            self.get_acc_phy_min()
            self.get_acc_phy_unit()
            self.get_acc_digital_max()
            self.get_acc_digital_min()
            self._acc_phy_range = self._acc_phy_max - self._acc_phy_min
            self._acc_dig_range = self._acc_dig_max - self._acc_dig_min
            
            logger.debug(self)
            self._register_callback()
            return True
        else:
            logger.error(f"设备[{self._box_mac}]连接失败")
            return False
    
    def get_eeg_phy_max(self):
        if self._handle:
            try:
                self._eeg_phy_max = _dll.ar4_sdk_get_eeg_phy_max(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取eeg物理最大值异常: {str(e)}")
        
        return self._eeg_phy_max
    
    def get_eeg_phy_min(self):
        if self._handle:
            try:
                self._eeg_phy_min = _dll.ar4_sdk_get_eeg_phy_min(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取eeg物理最小值异常: {str(e)}")
        
        return self._eeg_phy_min

    def get_eeg_phy_unit(self):
        if self._handle:
            try:
                eeg_unit = _dll.ar4_sdk_get_eeg_phy_unit(self._handle)
                if eeg_unit:
                        self._eeg_phy_unit = eeg_unit.decode("utf-8")
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取eeg物理单位异常: {str(e)}")
        
        return self._eeg_phy_unit
    
    def get_eeg_digital_max(self):
        if self._handle:
            try:
                self._eeg_dig_max = _dll.ar4_sdk_get_eeg_digtal_max(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取eeg数字最大值异常: {str(e)}")
        
        return self._eeg_dig_max
    
    def get_eeg_digital_min(self):
        if self._handle:
            try:
                self._eeg_dig_min = _dll.ar4_sdk_get_eeg_digtal_min(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取eeg数字最小值异常: {str(e)}")
        
        return self._eeg_dig_min
    
    
    def get_acc_phy_max(self):
        if self._handle:
            try:
                self._acc_phy_max = _dll.ar4_sdk_get_acc_phy_max(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取acc物理最大值异常: {str(e)}")
        
        return self._acc_phy_max
    
    def get_acc_phy_min(self):
        if self._handle:
            try:
                self._acc_phy_min = _dll.ar4_sdk_get_acc_phy_min(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取acc物理最小值异常: {str(e)}")
        
        return self._eeg_phy_min

    def get_acc_phy_unit(self):
        if self._handle:
            try:
                acc_unit = _dll.ar4_sdk_get_acc_phy_unit(self._handle)
                if acc_unit:
                        self._acc_phy_unit = acc_unit.decode("utf-8")
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取acc物理单位异常: {str(e)}")
        
        return self._acc_phy_unit
    
    def get_acc_digital_max(self):
        if self._handle:
            try:
                self._acc_dig_max = _dll.ar4_sdk_get_acc_digtal_max(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取acc数字最大值异常: {str(e)}")
        
        return self._acc_dig_max
    
    def get_acc_digital_min(self):
        if self._handle:
            try:
                self._acc_dig_min = _dll.ar4_sdk_get_acc_digtal_min(self._handle)
            except Exception as e:
                logger.error(f"设备[{self._box_mac}]获取acc数字最小值异常: {str(e)}")
        
        return self._acc_dig_min
    
    def _register_callback(self):
        try:
            _dll.ar4_sdk_register_data_notify(self._handle, self._data_callback)
            _dll.ar4_sdk_register_trigger_notify(self._handle, self._trigger_callback)
            _dll.ar4_sdk_register_conn_notify(self._handle, self._connected_notify_callback, self._disconnected_notify_callback)
            _dll.ar4_sdk_register_record_state_notify(self._handle, self._start_record_notify_callback, self._stop_record_notify_callback)
            _dll.ar4_sdk_register_box_conn_notify(self._handle, self._box_connected_notify_callback, self._box_disconnected_notify_callback)
        except Exception as e:
            logger.error(f"设备[{self._box_mac}]回调函数注册异常: {str(e)}")

    def update_info(self):
        self._last_time = _get_time()
        
    # 设备连接
    def connect(self)-> bool:
        
        if not self._box_mac:
            raise Exception("设备MAC地址不能为空")
        
        try:
            self._handle = _dll.ar4_sdk_connect(int(self._box_mac, 16), c_void_p(0))
            # logger.info(f"conn handle is {self._handle}")
            self._connected = self._handle is not None
            if self._handle:
                self._connected = True
                self._handle = c_void_p(self._handle)     
                self._conn_time = _get_time()
                self.get_box_name()
                self.get_head_mac()
                self.get_record_conn_state()
                self._last_time = _get_time()
            else:
                raise Exception(f"设备 {self._box_mac} 连接失败: {self._handle}")
            logger.debug(f"ar4 {self._box_mac} 连接: {self._handle}")                   
        except Exception as e:
            logger.error(f"ar4 {self._box_mac} 连接异常: {str(e)}")
            
        return self._connected
    
    # 读取盒子名称
    def get_box_name(self):
        if self._handle:    
            try:        
                self._head_mac = _dll.ar4_sdk_get_recoder_mac(self._handle)
            except Exception as e:
                    logger.error(f"ar4 {self._box_mac} 获取盒子记录子mac异常: {str(e)}")
    # 读取记录子mac
    def get_head_mac(self):
        if self._handle:    
            try:        
                self._box_name = _dll.ar4_sdk_get_record_conn_state(self._handle)
            except Exception as e:
                    logger.error(f"ar4 {self._box_mac} 获取盒子名称异常: {str(e)}")
    # 读取记录子连接状态
    def get_record_conn_state(self):
        if self._handle:    
            try:        
                self._head_conn_state = _dll.ar4_sdk_get_record_conn_state(self._handle)
            except Exception as e:
                    logger.error(f"ar4 {self._box_mac} 获取盒子记录子连接状态异常: {str(e)}")

    # 数据采集启动
    def start_acquisition(self):
        logger.info(f"device {self._box_mac} 启动数据采集...")
        if not self._handle:
            self.init()
        
        self._acq_info["start_time"] = _get_time()
        
        if self._handle:
            # 启动采集
            try:
                logger.debug(f"device {self._box_mac} 启动采集: {self._handle}")
                ret = _dll.ar4_sdk_start_acq(self._handle)
                if ret == 0:
                    logger.info(f"device {self._box_mac} 启动数据采集成功")
                    self.start_record()
                else:
                    logger.error(f"device {self._box_mac} 启动数据采集失败, ret: {ret}")
                return ret == 0
            except Exception as e:
                logger.error(f"device {self._box_mac} 停止采集异常: {str(e)}")
        else:
            logger.info(f"device {self._box_mac} 启动数据采集失败， 设备未连接")
            return False

    def stop_acquisition(self):
        """停止采集"""
        if self._handle:
            try:                
                self.get_acq_start_time() 
            except Exception as e:
                logger.error(f"ar4 {self._box_mac} 获取开始采集时间异常: {str(e)}")
            try:
                ret = _dll.ar4_sdk_stop_acq(self._handle)
                if ret == 0:
                    logger.info(f"device {self._box_mac} 停止采集成功")
                    self.stop_record()
                else:
                    logger.error(f"device {self._box_mac} 停止采集失败, ret: {ret}")
                    
                return ret == 0
            except Exception as e:
                logger.error(f"ar4 {self._box_mac} 停止采集异常: {str(e)}")
        else:
            return False
        
    def start_record(self):
        pass
    
    def stop_record(self):
        pass
    
    def trigger(self, trigger_type: int):
        pass

    def set_storage_path(self, storage_path):
        pass
    
    def disconnect(self):
        """断开连接"""
        if self._handle:
            _dll.ar4_sdk_disconnect(self._handle)

    def get_sample_rate(self):
        try:
            ret = _dll.ar4_sdk_get_record_sample_rate(self._handle)
            logger.debug(f"ar4 {self._box_mac} 获取采样率: {ret}")
            if ret > 1:
                self._sample_frequency = ret
        except Exception as e:
            logger.error(f"ar4 {self._box_mac} 获取采样率异常: {str(e)}")
            
    def get_acq_start_time(self):
        try:
            ret = _dll.ar4_sdk_get_acq_start_time(self._handle)
            # 更新采样开始时间
            if ret:
                self._acq_info["start_time"] = ret
            logger.debug(f"ar4 {self._box_mac} 获取采样开始时间: {ret}")
        except Exception as e:
            logger.error(f"ar4 {self._box_mac} 获取采样开始时间异常: {str(e)}")
    
    def _wrap_data_accept(self):
        
        @FuncAr4DataNotify
        def data_accept(handle, data_ptr):
            self._data_accept(data_ptr)
            
        return data_accept
    def _data_accept(self, data_ptr):
        self.eeg_accept(LMPacket().transfer(data_ptr.contents))
    
    def eeg_accept(self, packet: LMPacket):  
        pass
    
    def eeg2phy(self, digital:int):
        # 向量化计算（自动支持广播）
        return ((digital - self._eeg_dig_min) / self._eeg_dig_range) * self._eeg_phy_range + self._eeg_phy_min
    
    def acc2phy(self, digital:int):
        return ((digital - self._acc_dig_min) / self._acc_dig_range) * self._acc_phy_range + self._acc_phy_min
    
    def _wrap_trigger_accept(self):
        
        @FuncAr4DataNotify
        def trigger_accept(handle, time_ms, trigger_type, trigger_value):
            self._trigger_accept(time_ms, trigger_type, trigger_value)
            
        return trigger_accept
        
    def _trigger_accept(self, time_ms, trigger_type, trigger_value):
        logger.info(f"_trigger_accept 被调用")
        logger.info(f"触发时间: {time_ms}, 触发类型: {trigger_type}, 触发值: {trigger_value}")
        
        
    def _wrap_connected_notify(self):
        
        @FuncAr4RecorderConnected
        def connected_notify(handle):
            self._connected_notify(handle)
            
        return connected_notify      
    
    def _connected_notify(self, handle):
        logger.info(f"_connected_notify 被调用 handle: {handle}")
        
        
    def _wrap_disconnected_notify(self):
        @FuncAr4RecorderDisconnected
        def disconnected_notify(handle):
            self._disconnected_notify(handle)
            
        return disconnected_notify      
    
    def _disconnected_notify(self, handle):
        logger.info(f"_disconnected_notify 被调用 handle: {handle}")
        
        
    def _wrap_box_connected_notify(self):
        
        @FuncAr4RecorderConnected
        def box_connected_notify(handle):
            self._box_connected_notify(handle)
            
        return box_connected_notify      
    
    def _box_connected_notify(self, handle):
        logger.info(f"_box_connected_notify 被调用 handle: {handle}")
        
    def _wrap_box_disconnected_notify(self):
        @FuncAr4RecorderDisconnected
        def box_disconnected_notify(handle):
            self._box_disconnected_notify(handle)
            
        return box_disconnected_notify      
    
    def _box_disconnected_notify(self, handle):
        logger.info(f"_box_disconnected_notify 被调用 handle: {handle}")
        
    def _wrap_start_record_notify(self):
        
        @FuncAr4RecorderConnected
        def start_record_notify(handle):
            self._start_record_notify(handle)
            
        return start_record_notify      
    
    def _start_record_notify(self, handle):
        logger.info(f"_start_record_notify 被调用 handle: {handle}")        
        self._recording = True
        self._record_start_time = time()
        logger.info(self)        
        
    def _wrap_stop_record_notify(self):
        @FuncAr4RecorderDisconnected
        def stop_record_notify(handle):
            self._stop_record_notify(handle)
            
        return stop_record_notify      
    
    def _stop_record_notify(self, handle):
        logger.info(f"_stop_record_notify 被调用 handle: {handle}")      
        self._recording = False
        
    def __str__(self):
        return f"""
                box mac: {self._box_mac}, 
                box name: {self._box_name}, 
                box soc: {self._box_soc}
                head conn state: {self._head_conn_state}
                head mac: {self._head_mac},
                head soc: {self._head_soc} 
                connected: {self._connected}
                connect time: {self._conn_time}
                last time: {self._last_time}
                [
                eeg phy max: {self._eeg_phy_max}
                eeg phy min: {self._eeg_phy_min}
                eeg dig max: {self._eeg_dig_max}
                eeg dig min: {self._eeg_dig_min}
                eeg phy unit: {self._eeg_phy_unit}
                ]
                [
                acc phy max: {self._acc_phy_max}
                acc phy min: {self._acc_phy_min}
                acc dig max: {self._acc_dig_max}
                acc dig min: {self._acc_dig_min}
                acc phy unit: {self._acc_phy_unit}
                ]
                dig -> ((dig - {self._eeg_dig_min}) / {self._eeg_dig_range}) * {self._eeg_phy_range} + {self._eeg_phy_min}
                dig -> ((dig - {self._acc_dig_min}) / {self._acc_dig_range}) * {self._acc_phy_range} + {self._acc_phy_min}
            """
     


from multiprocessing import Queue
from typing import Literal
from loguru import logger
from qlsdk.persist import ARSKindlingEDFHandler
from qlsdk.rsc.interface import IDevice
from qlsdk.rsc.command import *
from qlsdk.rsc.device.base import QLBaseDevice

    
def intersection_positions(A, B):
    setB = set(B)
    seen = set()
    return [idx for idx, elem in enumerate(A) 
            if elem in setB and elem not in seen and not seen.add(elem)]
 
def get_sorted_indices_basic(A, B):
    """
    找出数组A中存在于数组B的元素，并返回这些元素在A中的位置，且顺序与它们在B中的顺序一致。

    参数:
        A (list): 待查找的数组。
        B (list): 作为参考的数组。

    返回:
        list: 一个列表，包含符合条件的元素在A中的索引，顺序与这些元素在B中的出现顺序一致。
    """
    # 1. 创建B中元素到其索引的映射
    b_index_map = {value: idx for idx, value in enumerate(B)}
    
    # 2. 筛选A中存在于B的元素，并记录其在A中的索引及在B中的位置
    # 使用列表推导式，同时避免重复元素干扰（如果B有重复，以第一次出现为准）
    found_elements = []
    for idx_a, value in enumerate(A):
        if value in b_index_map:
            # 记录: (该元素在A中的索引, 该元素在B中的索引)
            found_elements.append((idx_a, b_index_map[value]))
    
    # 3. 根据元素在B中的位置进行排序
    # 排序的依据是元组的第二个元素，即 b_index_map[value]
    found_elements_sorted = sorted(found_elements, key=lambda x: x[1])
    
    # 4. 提取排序后在A中的索引，形成结果数组C
    array_C = [item[0] for item in found_elements_sorted]
    
    return array_C
   
class ARSKindling(QLBaseDevice):
    
    device_type = 0x60 # C64RS设备类型标识符
    
    def __init__(self, socket):
        super().__init__(socket)
        
        self.channel_mapping = {
            "A1": 55, "A2": 56, "A3": 53, "A4": 54, "A5": 51, "A6": 52, "A7": 49, "A8": 50,
            "A9": 57, "A10": 58, "A11": 59, "A12": 60, "A13": 61, "A14": 62, 
            
            "B1": 39, "B2": 40,  "B3": 37, "B4": 38, "B5": 35, "B6": 36, "B7": 33, "B8": 34,
            "B9": 41, "B10": 42, "B11": 43, "B12": 44, "B13": 45, "B14": 46, 
            
            "C1": 23, "C2": 24, "C3": 21, "C4": 22, "C5": 19, "C6": 20, "C7": 17, "C8": 18,
            "C9": 25, "C10": 26, "C11": 27, "C12": 28, "C13": 29, "C14": 30, 
            
            "D1": 7, "D2": 8, "D3": 5, "D4": 6, "D5": 3, "D6": 4, "D7": 1, "D8": 2,
            "D9": 9, "D10": 10, "D11": 11, "D12": 12, "D13": 13, "D14": 14, 
        }
        
        self._channel_spilt = {
            "A" : [55, 56, 53, 54, 51, 52, 49, 50, 57, 58, 59, 60, 61, 62],
            "B" : [39, 40, 37, 38, 35, 36, 33, 34, 41, 42, 43, 44, 45, 46],
            "C" : [23, 24, 21, 22, 19, 20, 17, 18, 25, 26, 27, 28, 29, 30],
            "D" : [7, 8, 5, 6, 3, 4, 1, 2, 9, 10, 11, 12, 13, 14],
        }
        
        self._channel_mapping = {
            1: "7", 
            2: "8",
            3: "5", 
            4: "6",
            5: "3", 
            6: "4",
            7: "1", 
            8: "2",
            9: "9", 
            10: "10",
            11: "11", 
            12: "12",
            13: "13", 
            14: "14",
            
            17: "7", 
            18: "8",
            19: "5", 
            20: "6",
            21: "3", 
            22: "4",
            23: "1", 
            24: "2",
            25: "9", 
            26: "10",
            27: "11", 
            28: "12",
            29: "13", 
            30: "14",
            
            33: "7", 
            34: "8",
            35: "5", 
            36: "6",
            37: "3", 
            38: "4",
            39: "1", 
            40: "2",
            41: "9", 
            42: "10",
            43: "11", 
            44: "12",
            45: "13", 
            46: "14",
            
            49: "7", 
            50: "8",
            51: "5", 
            52: "6",
            53: "3", 
            54: "4",
            55: "1", 
            56: "2",
            57: "9", 
            58: "10",
            59: "11", 
            60: "12",
            61: "13", 
            62: "14"
        }
                
    @classmethod
    def from_parent(cls, parent:IDevice) -> IDevice:
        rlt = cls(parent.socket)
        rlt.device_id = parent.device_id
        rlt._device_no = parent.device_no
        return rlt
        
    def init_edf_handler(self):
        self._edf_handler = ARSKindlingEDFHandler(self.sample_rate,  self.sample_range * 1000 , - self.sample_range * 1000, self.resolution)  
        self._edf_handler.set_device_type(self.device_type)
        self._edf_handler.set_device_no(self.device_no)
        self._edf_handler.set_storage_path(self._storage_path)
        self._edf_handler.set_file_prefix(self._file_prefix if self._file_prefix else 'ARS')
        logger.debug(f"EDF Handler initialized")
        
    # 设置采集参数
    def set_acq_param(self, channels, sample_rate = 500, sample_range = 188):
        self._acq_param["original_channels"] = channels
        
        # 根据映射关系做通道转换
        for k in channels.keys():
            if isinstance(channels[k], list):
                temp = [k + str(i) for i in channels[k]]
                channels[k] = [self.channel_mapping.get(c, 1) for c in temp]
            else:
                channels[k] = [k + str(channels[k])]
                
                
        # 更新采集参数        
        self._acq_param["channels"] = channels 
        self._acq_param["sample_rate"] = sample_rate
        self._acq_param["sample_range"] = sample_range
        self._acq_channels = channels
        self._sample_rate = sample_rate
        self._sample_range = sample_range     
    
    # 设置阻抗通道
    # {'A': [1,2,3], 'B':[1,2,3]}
    def set_impedance_channels(self, channels):
        
        arr = []
        
        # 根据映射关系做通道转换
        for k in channels.keys():
            if isinstance(channels[k], list):
                temp = [k + str(i) for i in channels[k]]
                arr += [self.channel_mapping.get(c, 1) for c in temp]
            else:
                arr += [k + str(channels[k])]
                
        self._impedance_channels = arr
        
    @property
    def acq_channels(self):
        if self._acq_channels is None:
            self._acq_channels = [i for i in range(1, 63)]
            
        arr = []
        for k in self._acq_channels.keys():
            arr = list(arr + self._acq_channels[k])
            
        return list(set(arr))    
    
    def _produce_impedance(self, body: bytes):
        # 分发阻抗数据包给订阅者
        if len(self._impedance_consumer) > 0:
            packet = self._impedance_wrapper(body)
            real_data = self.__impedance_transfer(packet)
            for topic, q in self._impedance_consumer.items():
                try:
                    # 队列满了就丢弃最早的数据
                    if q.full(): 
                        q.get()
                    q.put(real_data, timeout=1)
                except Exception as e:
                    logger.error(f"impedance data put to queue exception: {str(e)}")
                    
    def _produce_signal(self, body: bytes):        
        
        # 处理信号数据
        data = self._signal_wrapper(body)  
        # logger.debug("pkg_id: {}, eeg len: {}".format(data.pkg_id, len(data.eeg)))
        
        trigger_positions = [index for index, value in enumerate(data.trigger) if value != 0]
        if len(trigger_positions) > 0:
            # logger.debug(f"Trigger触发点位置: {trigger_positions}, 触发点时间戳: {[data.time_stamp + int(pos * 1000 / data.sample_rate) for pos in trigger_positions]}")
            for pos in trigger_positions:
                self.trigger(self.trigger_info(data.trigger[pos]))
        
        if len(self.signal_consumers) > 0 :
            # 信号数字值转物理值
            data.eeg_p = self.eeg2phy(np.array(data.eeg))        
        
        real_data = self.__signal_transfer(data)   
               
        # 存储
        if self.storage_enable:            
            # 确保记录线程启动
            if self._recording is False:
                self._start_recording() 
                
            # 写入文件的缓存队列
            if self._signal_cache is None:
                self._signal_cache = Queue(1000000)  # 缓冲队列
            tmp = real_data.copy()
            self._signal_cache.put(tmp)
                                    
        if len(self.signal_consumers) > 0 :
            # 订阅只返回物理值
            real_data.eeg = real_data.eeg_p
            # 发送数据包到订阅者
            for q in list(self.signal_consumers.values()):
                # 队列满了就丢弃最早的数据
                if q.full(): 
                    q.get()
                    
                q.put(real_data)      
     
    # 信号数据转换
    def __impedance_transfer(self, packet: ImpedancePacket):
        channels = {}
        impedance = {}
        #按分区拆分数据格式
        for k in self._channel_spilt.keys():      
            logger.trace(f'分区{k}的阻抗数据')   
            c, d, p = self.__packet_filter(packet, self._channel_spilt[k], type='impedance')
            if c is not None:
                channels[k] = c
                impedance[k] = d
        packet.channels = channels
        packet.impedance = impedance
            
        return packet
      
    # 信号数据转换
    def __signal_transfer(self, packet: RscPacket):
        channels = {}
        eeg = {}
        eeg_p = {}
        #按分区拆分数据格式
        for k in self._channel_spilt.keys():      
            logger.trace(f'分区{k}, {self._channel_spilt[k]}')   
            c, d, p = self.__packet_filter(packet, self._channel_spilt[k])
            if c is not None:
                channels[k] = c
                eeg[k] = d
            if p is not None:
                eeg_p[k] = p
        packet.channels = channels
        packet.eeg = eeg
        #物理值
        if packet.eeg_p is not None:
            packet.eeg_p = eeg_p
            
        return packet
    
    def __packet_filter(self, packet, channel_filter=None, type:Literal['signal','impedance']='signal'):
        # 参数检查
        if packet is None or channel_filter is None:
            logger.warning("空数据，忽略")
            return None, None, None
        
        channel_pos = get_sorted_indices_basic(packet.channels, channel_filter)
        
        if channel_pos is None or len(channel_pos) == 0 :
            logger.trace(f"没有指定分区的通道，跳过")
            return None, None, None
        
        # 分区数据包写入
        
        # 保留本分区的通道和数据
        channels = [packet.channels[p] for p in channel_pos]
        
        #阻抗数据
        if type == 'impedance':
            impedance = [packet.impedance[p] for p in channel_pos]
            return [self._channel_mapping.get(channel_id, []) for channel_id in channels], impedance, None
        
        # 信号数据
        eeg = [packet.eeg[p] for p in channel_pos]
        eeg_p = None
        if packet.eeg_p is not None:
            eeg_p = [packet.eeg_p[p] for p in channel_pos]
        
        return [self._channel_mapping.get(channel_id, []) for channel_id in channels], eeg, eeg_p
        
    
    


import socket
from typing import Optional
from loguru import logger   
from qlsdk.rsc.network.discover import UdpBroadcaster
from threading import Thread
from qlsdk.core import *
from qlsdk.rsc.device import QLBaseDevice, DeviceFactory
from qlsdk.rsc.interface import IDevice
from qlsdk.rsc.proxy import DeviceProxy
import time
        
        
class DeviceContainer(object):
    def __init__(self, proxy_enabled=False, tcp_port = 19216):
        self._devices = {}
        self._tcp_port = tcp_port
        self._proxy_enabled = proxy_enabled
        
        # 设备搜索广播器
        self._broadcaster = UdpBroadcaster()
        self._broadcaster.start()
        
        # 监听设备连接
        self._listening_thread = Thread(target=self._listening)
        self._listening_thread.daemon = True
        self._listening_thread.start()
        
    @property
    def devices(self)-> IDevice:
        return self._devices
    
    '''
        等待设备连接
    '''
    def connect(self, device_id: str, timeout:int=30) -> Optional[IDevice]:        
        logger.info(f"开始搜索设备: {device_id}")
        self.add_search(device_id)
        for s in range(timeout):
            device = self.get_device(device_id)
            if device:
                logger.success(f"设备[{device_id}]已连接成功。")
                return device
            time.sleep(1)
        logger.warning(f"在{timeout}内未搜索到设备：{device_id}")
        return None
            
    def _listening(self):
        tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            # 绑定到所有接口的19216端口
            tcp_socket.bind(('0.0.0.0', self._tcp_port))
            tcp_socket.listen(100)
            logger.info(f"端口[{self._tcp_port}]监听服务已启动")

            while True:
                client_socket, addr = tcp_socket.accept()
                logger.info(f"接收到来自[{addr[0]}:{addr[1]}]的连接，待确认设备类型...")
                
                
                # 为每个新连接创建线程处理
                try:
                    client_handler = Thread(
                        target=self.client_handler,
                        args=(client_socket,),
                        daemon=False
                    )
                    
                    client_handler.start()
                except Exception as e:
                    logger.error(f"Handler Error: {e}")
                
        except KeyboardInterrupt:
            logger.error(f"端口[{self._tcp_port}]监听服务异常关闭")
        except Exception as e:
            logger.error(f"端口[{self._tcp_port}]监听服务异常: {e}")
        finally:
            logger.error(f"端口[{self._tcp_port}]监听服务关闭")
            tcp_socket.close()
            
    def client_handler(self, client_socket):
                
        if self._proxy_enabled:
            # 启动代理  TODO: 代理的同时支持接口控制和数据转发
            proxy = DeviceProxy(client_socket)
            proxy.start()
        else:            
            # 数据监听
            device = QLBaseDevice(client_socket)
            device.start_listening()
            # GET_DEVICE_INFO
            msg = GetDeviceInfoCommand.build(device).pack()
            logger.debug(f"发送获取设备信息命令: {msg.hex()}")
            device.send(msg)
            # 添加设备
            while True:
                if device.device_no:
                    real_device = DeviceFactory.create_device(device)
                    device.stop_listening()
                    real_device.start_listening()
                    self.add_device(real_device)
                    logger.info(f"设备[{device.device_name}]已连接，设备类型为：{hex(real_device.device_type)}")
                    break
            
        
    # 搜寻设备
    def add_search(self, device_id):
        self._broadcaster.add_device(device_id)
        
    def add_device(self, device: IDevice):
        if device is None or device.device_no is None:
            logger.warning("无效的设备")
            
        self._devices[device.device_no] = device
        logger.info(f"添加设备[{device.device_no}]到设备列表，已连接设备数量：{len(self._devices)}")
        
        # 标记设备为已连接
        self._broadcaster.mark_device_as_connected(device.device_no)
        
    def remove_device(self, device_no:str=None):
        if device_no is None:
            return
        
        if device_no in self._devices:
            del self._devices[device_no]
            logger.info(f"从设备列表移除设备[{device_no}]，已连接设备数量：{len(self._devices)}")
        
    def get_device(self, device_no:str=None)->IDevice:
        logger.info(f"已连接设备数量:{len(self._devices)}")
        if len(self._devices) == 0:
            return None
        
        # 未指定device_id，返回第一个设备
        if device_no is None:
            return list(self._devices.values())[0]
        
        return self._devices.get(device_no, None)
            


from datetime import datetime
from multiprocessing import Lock, Queue
from pyedflib import FILETYPE_BDFPLUS, FILETYPE_EDFPLUS, EdfWriter
from threading import Thread
from loguru import logger
import numpy as np
import os

class EdfHandler(object):
    def __init__(self, sample_frequency, physical_max, physical_min, digital_max, digital_min, resolution=16, storage_path = None):
        self.physical_max = physical_max
        self.physical_min = physical_min
        self.digital_max = digital_max
        self.digital_min = digital_min
        self.eeg_channels = None
        self.eeg_sample_rate = 500
        self.acc_channels = None
        self.acc_sample_rate = 50
        self._cache = Queue()
        self.resolution = resolution
        self.sample_frequency = sample_frequency
        # bytes per second
        self.bytes_per_second = 0
        self._edf_writer = None
        self._cache2 = tuple()
        self._recording = False
        self._edf_writer = None
        self.annotations = None
        # 每个数据块大小
        self._chunk = np.array([])
        self._Lock = Lock()
        self._duration = 0
        self._points = 0
        self._first_pkg_id = None
        self._last_pkg_id = None
        self._first_timestamp = None
        self._first_time = None
        self._end_time = None
        self._patient_code = "patient_code"
        self._patient_name = "patient_name"
        self._device_type = None
        self._total_packets = 0
        self._lost_packets = 0
        self._storage_path = storage_path
        
    @property
    def file_name(self): 
        if self._storage_path:
            try:
                os.makedirs(self._storage_path, exist_ok=True)  # 自动创建目录，存在则忽略
                return f"{self._storage_path}/{self._device_type}_{self._first_timestamp}.edf"
            except Exception as e:
                logger.error(f"创建目录[{self._storage_path}]失败: {e}")
            
        return f"{self._device_type}_{self._first_timestamp}.edf"
         
    @property
    def file_type(self):
        return FILETYPE_BDFPLUS if self.resolution == 24 else FILETYPE_EDFPLUS
    
    def set_device_type(self, device_type):
        self._device_type = device_type
        
    def set_storage_path(self, storage_path):
        self._storage_path = storage_path
        
    def set_patient_code(self, patient_code):
        self._patient_code = patient_code
        
    def set_patient_name(self, patient_name):
        self._patient_name = patient_name
    
    def append(self, data):
        if data:
            # 通道数
            if self._first_pkg_id is None:
                self.eeg_channels = data.eeg_ch_count
                self.acc_channels = data.acc_ch_count
                self._first_pkg_id = data.pkg_id
                self._first_timestamp = data.time_stamp
                self._first_time = datetime.now()
                
            if self._last_pkg_id and self._last_pkg_id != data.pkg_id - 1:  
                self._lost_packets += data.pkg_id - self._last_pkg_id - 1
                logger.warning(f"数据包丢失: {self._last_pkg_id} -> {data.pkg_id}, 丢包数: {data.pkg_id - self._last_pkg_id - 1}")
                
            self._last_pkg_id = data.pkg_id
            self._total_packets += 1
            
        # 数据
        self._cache.put(data)
        if not self._recording:
            self.start()
    
    # trigger标记
    # desc: 标记内容
    # cur_time: 设备时间时间戳，非设备发出的trigger不要设置
    def trigger(self, desc, cur_time=None):
        if self._edf_writer is None:
            logger.warning("EDF writer未初始化，无法写入trigger标记")
            return
        
        if cur_time is None:
            onset = datetime.now().timestamp() - self._first_time.timestamp()
        else:
            onset = cur_time - self._first_timestamp
            
        self._edf_writer.writeAnnotation(onset, 1, desc)
    
    
    # def trigger(self, desc):
    #     if self._edf_writer:            
    #         self._edf_writer.writeAnnotation(0, 1, desc)
    
    def start(self):
        self._recording = True
        record_thread = Thread(target=self._consumer)
        record_thread.start()
        
    def _consumer(self):
        logger.debug(f"开始消费数据 _consumer: {self._cache.qsize()}")
        while True:
            if self._recording or (not self._cache.empty()):
                try:
                    data = self._cache.get(timeout=10)
                    if data is None:
                        break
                    # 处理数据
                    self._points += len(data.eeg[0])
                    self._write_file(data.eeg, data.acc)
                except Exception as e:
                    logger.error("数据队列为空，超时(10s)结束")
                    break
            else:
                break
            
        self.close()
        
    def _write_file(self, eeg_data, acc_data):
        try:
            if self._edf_writer is None:
                self.initialize_edf()
            
            if (self._chunk.size == 0):
                self._chunk = np.asarray(eeg_data)
            else:                
                self._chunk = np.hstack((self._chunk, eeg_data))
                
            if self._chunk.size >= self.eeg_sample_rate * self.eeg_channels:                
                self._write_chunk(self._chunk[:self.sample_frequency])
                self._chunk = self._chunk[self.sample_frequency:]            
            
        except Exception as e:
            logger.error(f"写入数据异常: {str(e)}")
        
    def close(self):
        self._recording = False
        if self._edf_writer:            
            self._end_time = datetime.now().timestamp()
            self._edf_writer.writeAnnotation(0, 1, "start recording ")
            self._edf_writer.writeAnnotation(self._duration, 1, "recording end")
            self._edf_writer.close()
        
        logger.info(f"文件: {self.file_name}完成记录, 总点数: {self._points}, 总时长: {self._duration}秒 丢包数: {self._lost_packets}/{self._total_packets + self._lost_packets}")
        
    
        
    def initialize_edf(self):
        # 创建EDF+写入器
        self._edf_writer = EdfWriter(
            self.file_name,
            self.eeg_channels,
            file_type=self.file_type
        )
        
        # 设置头信息
        self._edf_writer.setPatientCode(self._patient_code)
        self._edf_writer.setPatientName(self._patient_name)
        self._edf_writer.setEquipment(self._device_type)
        self._edf_writer.setStartdatetime(datetime.now())
        
        # 配置通道参数
        signal_headers = []
        for ch in range(self.eeg_channels):
            signal_headers.append({
                "label": f'channels {ch + 1}',
                "dimension": 'uV',
                "sample_frequency": self.sample_frequency,
                "physical_min": self.physical_min,
                "physical_max": self.physical_max,
                "digital_min": self.digital_min,
                "digital_max": self.digital_max
            })
        
        self._edf_writer.setSignalHeaders(signal_headers)
    
    def _write_chunk(self, chunk):
        logger.debug(f"写入数据: {chunk}")
        # 转换数据类型为float64（pyedflib要求）
        data_float64 = chunk.astype(np.float64)
        # 写入时转置为(样本数, 通道数)格式
        self._edf_writer.writeSamples(data_float64)
        self._duration += 1

    


import abc
import numpy as np
from time import time_ns
from typing import Dict, Type
from loguru import logger


from qlsdk.core.crc import crc16
from qlsdk.core.entity import RscPacket, ImpedancePacket
from qlsdk.core.utils import to_channels, to_bytes
from qlsdk.rsc.interface import IDevice

class DeviceCommand(abc.ABC):
    # 消息头
    HEADER_PREFIX = b'\x5A\xA5'
    # 消息头总长度 2(prefix) +1(pkgType) +1(deviceType) +4(deviceId) +4(len) +2(cmd)
    HEADER_LEN = 14  
    # 消息指令码位置
    CMD_POS = 12
    
    def __init__(self, device: IDevice):
        self.device = device
        
    @classmethod
    def build(cls, device) :
        return cls(device)
    
    @property
    @abc.abstractmethod
    def cmd_code(self) -> int:
        pass
    @property
    @abc.abstractmethod
    def cmd_desc(self) -> str:
        pass
    
    @staticmethod
    def checksum(data: bytes) -> bytes:
        return crc16(data).to_bytes(2, 'little')

    def pack(self, body=b'') -> bytes:
        # header+body+checksum
        body = self.pack_body()
        header = self.pack_header(len(body))
        payload = header + body
        return payload + DeviceCommand.checksum(payload)
    def pack_body(self) -> bytes:
        """构建消息体"""
        return b''
    def pack_header(self, body_len: int) -> bytes:
        # device_id = int(self.device.device_id) if self.device and self.device.device_id else 0
        device_type = int(self.device.device_type) if self.device and self.device.device_type else 0
        device_id = bytes.fromhex(self.device.device_id)[::-1] if self.device.device_id else bytes.fromhex('00000000')
        
        #兼容设计
        b_device_type = None
        if device_type > 0xFF:
            b_device_type = int.to_bytes(device_type, 2, 'little')
        
        if b_device_type is None:
            return (
                DeviceCommand.HEADER_PREFIX
                + int(2).to_bytes(1, 'little')     # pkgType  
                + device_type.to_bytes(1, 'little')
                + device_id
                + (DeviceCommand.HEADER_LEN + body_len + 2).to_bytes(4, 'little')  # +1 for checksum
                + self.cmd_code.to_bytes(2, 'little')
            )
        else:
            return (
                DeviceCommand.HEADER_PREFIX
                + b_device_type[0].to_bytes(1, 'little')     # pkgType  
                + b_device_type[1].to_bytes(1, 'little')
                + device_id.to_bytes(4, 'little')
                + (DeviceCommand.HEADER_LEN + body_len + 2).to_bytes(4, 'little')  # +1 for checksum
                + self.cmd_code.to_bytes(2, 'little')
            )
    
    def unpack(self, payload: bytes) -> bytes:
        """解析消息体"""
        # 解析消息体
        body = payload[self.HEADER_LEN:-2]   
    
    def parse_body(self, body: bytes):
        time = int.from_bytes(body[0:8], 'little')
        # result - 1B
        result = body[8]
        logger.debug(f"[{time}]{self.cmd_desc}{'成功' if result == 0 else '失败'}")
        
    

class CommandFactory:
    """Registry for command implementations"""
    _commands: Dict[int, Type[DeviceCommand]] = {}
    
    @classmethod
    def register_command(cls, code: int, command: Type[DeviceCommand]):
        cls._commands[code] = command
    
    @classmethod
    def create_command(cls, code: int) -> Type[DeviceCommand]:
        logger.trace(f"Creating command for code: {hex(code)}")
        if code not in cls._commands:
            logger.warning(f"不支持的设备指令: {hex(code)}")
            return cls._commands[DefaultCommand.cmd_code]
        return cls._commands[code]

# =============================================================================
class DefaultCommand(DeviceCommand):    
    cmd_code = 0x00   
    cmd_desc = "未定义"
        
    def parse_body(self, body: bytes):
        # Response parsing example: 2 bytes version + 4 bytes serial
        logger.info(f"Received body len: {len(body)}")
    
class GetDeviceInfoCommand(DeviceCommand):
    cmd_code = 0x17
    cmd_desc = "设备信息"
    
    def parse_body(self, body: bytes):
        # time - 8B
        self.device.connect_time = int.from_bytes(body[0:8], 'little')
        self.device.current_time = self.device.connect_time
        # result - 1B
        result = body[8]
        # deviceId - 4B
        # self.device.device_id = int.from_bytes(body[9:13], 'big')
        self.device.device_id = body[9:13].hex()
        # deviceType - 4B
        device_type = int.from_bytes(body[13:17], 'little')
        self.device.set_device_type(device_type)
        # softVersion - 4B
        self.device.software_version = body[17:21].hex()
        # hardVersion - 4B
        self.device.hardware_version = body[21:25].hex()
        # deviceName - 16B
        self.device.device_name = body[25:41].decode('utf-8').rstrip('\x00')
        self.device.set_device_no(self.device.device_name)
        # flag - 4B
        flag = int.from_bytes(body[41:45], 'little')
        logger.debug(f"Received device info: {result}, {flag}, {self.device}")
        
        # 创建设备对象
        

# 握手
class HandshakeCommand(DeviceCommand):
    cmd_code = 0x01
    cmd_desc = "握手"
    
    def parse_body(self, body: bytes):
        logger.info(f"Received handshake response: {body.hex()}")

# 查询电量
class QueryBatteryCommand(DeviceCommand):
    cmd_code = 0x16
    cmd_desc = "电量信息"
    def parse_body(self, body: bytes):
        # time - 8b
        self.device.current_time = int.from_bytes(body[0:8], 'little')
        # result - 1b
        result = body[8]
        # 更新设备信息
        if result == 0:
            # voltage - 2b mV
            self.device.voltage = int.from_bytes(body[9:11], 'little')
            # soc - 1b
            self.device.battery_remain = body[11]
            # soh - 1b
            self.device.battery_total = body[12]
            # state - 1b
            # state = body[13]
            logger.trace(f"电量更新: {self.device}")
        else:
            logger.warning(f"QueryBatteryCommand message received but result is failed.")

# 设置采集参数
class SetAcquisitionParamCommand(DeviceCommand):
    cmd_code = 0x451
    cmd_desc = "设置信号采集参数"
    
    def pack_body(self): 

        return self.device.gen_set_acquirement_param()
        
# 启动采集
class StartAcquisitionCommand(DeviceCommand):
    cmd_code = 0x452
    cmd_desc = "启动信号采集"
    
    def pack_body(self):
        return bytes.fromhex('0000')

# 停止采集
class StopAcquisitionCommand(DeviceCommand):
    cmd_code = 0x453
    cmd_desc = "停止信号采集"
    
    def pack_body(self):
        return b''

        
# 设置阻抗采集参数
class SetImpedanceParamCommand(DeviceCommand):
    cmd_code = 0x411
    cmd_desc = "设置阻抗测量参数"
    def pack_body(self): 

        return self.device.gen_set_impedance_param()
    def parse_body(self, body):
        result = body[8]
        if result == 0x00:
            logger.info("阻抗测量参数设置成功")
        else:
            logger.warning(f"阻抗测量参数设置失败: {int(result)}")
        
# 启动阻抗测量
class StartImpedanceCommand(DeviceCommand):
    cmd_code = 0x412
    cmd_desc = "启动阻抗测量"
    def pack_body(self):
        body = bytes.fromhex('0000')
        body += to_bytes(self.device.get_impedance_channels())
        body += bytes.fromhex('0000000000000000') # 8字节占位符
        return body
    
    def parse_body(self, body):
        result = body[8]
        if result == 0x00:
            logger.info("阻抗测量启动成功")
        else:
            logger.warning(f"阻抗测量启动失败: {int(result)}")
        

# 停止阻抗测量
class StopImpedanceCommand(DeviceCommand):
    cmd_code = 0x413
    cmd_desc = "停止阻抗测量"
    
    def pack_body(self):
        return b''
            
# 启动刺激
class StartStimulationCommand(DeviceCommand):
    cmd_code = 0x48C
    cmd_desc = "启动刺激"
    def pack_body(self):
        return self.device.get_stim_param()
        # return bytes.fromhex('01000000000000008813000000000000010000000000000000000140420f00640064000000803f0000010000000000000000000000000000000000000000008813000000000000')
    def parse_body(self, body: bytes):
        # time - 8B
        time = int.from_bytes(body[0:8], 'little')
        # result - 1B
        result = body[8]
        # error_channel - 8B
        # error_channel= int.from_bytes(body[9:17], 'big')
        channels = to_channels(body[9:17])
        logger.info(f"通道 {channels} 刺激开始")
        if self.device:
            self.device.trigger(f"通道 {channels} 刺激开始")
        # error_type - 1B
        error_type = body[17]

# 停止刺激
class StopStimulationCommand(DeviceCommand):
    cmd_code = 0x488
    cmd_desc = "停止刺激"
     
# 启动刺激
class StopStimulationNotifyCommand(DeviceCommand):
    cmd_code = 0x48D
    cmd_desc = "停止刺激通知"
    def pack_body(self):
        return self.device.stim_paradigm.to_bytes()
        # return bytes.fromhex('01000000000000008813000000000000010000000000000000000140420f00640064000000803f0000010000000000000000000000000000000000000000008813000000000000')
    def parse_body(self, body: bytes):
        # time - 8B
        time = int.from_bytes(body[0:8], 'little')
        # result - 1B
        result = body[8]
        # error_channel - 8B
        # error_channel= int.from_bytes(body[9:17], 'big')
        channels = to_channels(body[9:17])
        logger.info(f"通道 {channels} 刺激结束")
        if self.device:
            self.device.trigger(f"通道 {channels} 刺激结束", time)
        # error_type - 1B
        error_type = body[17]   
# 刺激信息
class StimulationInfoCommand(DeviceCommand):
    cmd_code = 0x48e
    cmd_desc = "刺激告警信息"
    
    def parse_body(self, body: bytes):
        time = int.from_bytes(body[0:8], 'little')
        # result - 1B
        result = body[8]
        # error_channel - 8B
        channels = to_channels(body[9:17])
        # 保留位-8B
        # error_type - 1B
        err_type = body[17]
        # 特征位-4B
        # errType  = int.from_bytes(body[25:29], 'little')
        logger.warning(f"刺激告警信息[{err_type}]，通道 {channels} 刺激驱动不足")
        

# 阻抗数据
class ImpedanceDataCommand(DeviceCommand):
    cmd_code = 0x415
    cmd_desc = "阻抗数据"
    
    def parse_body(self, body: bytes):
        self.device.produce(body, type="impedance")

# 信号数据
class SignalDataCommand(DeviceCommand):
    cmd_code = 0x455
    cmd_desc = "信号数据"
    
    def unpack(self, payload):
        return super().unpack(payload)
    
    def parse_body(self, body: bytes):     
        # # 解析数据包
        # packet = RscPacket()
        # packet.transfer(body)  
        # 将数据包传递给设备
        self.device.produce(body)
                
            

# =============================================================================
# 指令实现类注册到指令工厂
# =============================================================================
CommandFactory.register_command(DefaultCommand.cmd_code, DefaultCommand)
CommandFactory.register_command(GetDeviceInfoCommand.cmd_code, GetDeviceInfoCommand)
CommandFactory.register_command(HandshakeCommand.cmd_code, HandshakeCommand)
CommandFactory.register_command(QueryBatteryCommand.cmd_code, QueryBatteryCommand)
CommandFactory.register_command(SetAcquisitionParamCommand.cmd_code, SetAcquisitionParamCommand)
CommandFactory.register_command(StartAcquisitionCommand.cmd_code, StartAcquisitionCommand)
CommandFactory.register_command(StopAcquisitionCommand.cmd_code, StopAcquisitionCommand)
CommandFactory.register_command(SetImpedanceParamCommand.cmd_code, SetImpedanceParamCommand)
CommandFactory.register_command(StartImpedanceCommand.cmd_code, StartImpedanceCommand)
CommandFactory.register_command(StopImpedanceCommand.cmd_code, StopImpedanceCommand)
CommandFactory.register_command(StartStimulationCommand.cmd_code, StartStimulationCommand)
CommandFactory.register_command(StimulationInfoCommand.cmd_code, StimulationInfoCommand)
CommandFactory.register_command(ImpedanceDataCommand.cmd_code, ImpedanceDataCommand)
CommandFactory.register_command(SignalDataCommand.cmd_code, SignalDataCommand)





from .network.discover import *

from .entity import DeviceParser, QLDevice
from .command import *
# from .device_manager import DeviceContainer
from .proxy import DeviceProxy
from .paradigm import *
from .manager import DeviceContainer
from .network import *



from threading import Thread
import psutil
import time

def get_active_ipv4():
    ips = []
    # 获取所有接口地址信息
    all_addrs = psutil.net_if_addrs()
    # 获取接口状态（是否处于UP状态）
    stats = psutil.net_if_stats()
    
    for name, addrs in all_addrs.items():
        # 检查接口是否启用
        if stats[name].isup:
            for addr in addrs:
                # 提取IPv4地址
                if addr.family == socket.AF_INET:
                    ips.append(addr.address)
    return ips
def monitor_up_interfaces(interval=2, callback=None):
    prev_status = {iface: psutil.net_if_stats()[iface].isup 
                  for iface in psutil.net_if_stats()}
    
    while True:
        current_stats = psutil.net_if_stats()
        for iface, stats in current_stats.items():
            current_up = stats.isup
            # 检测状态变化
            if current_up != prev_status.get(iface, None):
                if current_up:
                    print(f"[UP] 接口 {iface} 激活")
                    if callback: callback(iface, "UP")
                else:
                    print(f"[DOWN] 接口 {iface} 断开")
                    if callback: callback(iface, "DOWN")
                prev_status[iface] = current_up
        time.sleep(interval)

# 自定义回调函数示例
def notify(iface, status):
    if status == "UP":
        print(f"接口 {iface} 已激活")

# 启动监听
# monitor_up_interfaces(callback=notify)

monitor = Thread(target=monitor_up_interfaces, name="Network Status Monitor", args=(2, notify))
monitor.start()

import socket
def is_port_in_use(ip, port):
    with socket.socket() as s:
        return s.connect_ex((ip, port)) == 0
    
print(get_active_ipv4())